# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import schema_analyzer  # noqa: F401
import data_analyzer  # noqa: F401



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import ast
import re

EXOGENOUS_INFORMATION_SUGGESTION = (
    "If for the body change, there can be construed a (simple) transformation "
    "function, consider to modify the known transformations file."
)


TARGET_RECORDS = {"noupdate": True}

BODY_FIELD_TYPES = ["char", "text", "html", "binary"]


# fmt: off
KOWN_TRANSITIONS_DESCRIPTIONS = {
    # Note: (column or model) relocations (refactorings) are covered by schema changes, so they are not of a concern here
    'disappear'                             : {'description': "Record X will disappear."},
    'body_change'                           : {'description': "Field X of record Y will change."},
}

KOWN_TRANSITIONS_INSTRUCTIONS = {
    'disappear'                             : {'instruction': "Post migrate might delete this record."},
    'body_change'                           : {'instruction': "Post migrate might execute a string transformation on this body." + EXOGENOUS_INFORMATION_SUGGESTION},
}


contains_something_like_a_model_string_regex = re.compile(r'[a-z0-9_.]+')

KNOWN_SUSPICIOUS_REGEX_DESCRIPTIONS = {
    'suspect_model'                         : {'description': "Record X field Y contains something that looks like a model reference",     'check': lambda string: bool(contains_something_like_a_model_string_regex.search(string))},
    'suspect_python'                        : {'description': "Record X field Y contains valid python code",                               'check': lambda string: is_valid_python(string)},
}


def is_valid_python(code):
    try:
        ast.parse(code)
    except SyntaxError:
        return False
    return True

# fmt: on



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

from contextlib import contextmanager

import odoo
from odoo.api import Environment


@contextmanager
def OdooAnalyzerEnvironment(database):
    with Environment.manage():
        registry = odoo.registry(database)
        try:
            with registry.cursor() as cr:
                uid = odoo.SUPERUSER_ID
                ctx = Environment(cr, uid, {})["res.users"].context_get()
                env = Environment(cr, uid, ctx)
                cr.rollback()
                yield env
                cr.rollback()
        finally:
            odoo.modules.registry.Registry.delete(database)
            odoo.sql_db.close_db(database)



# -*- coding: utf-8 -*-
# Copyright 2016-2017 Camptocamp SA
# Copyright 2017-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

from __future__ import print_function

import datetime
import logging
import os
import sys

import semver
import yaml
from dodoo import odoo

from .database import MigrationTable
from .exceptions import MigrationErrorGap, MigrationErrorUnfinished, ParseError

if odoo.release.version_info[0] > 10:
    from odoo import migration
else:
    migration = None

PY3 = sys.version_info[0] == 3
string_type = str if PY3 else basestring  # noqa

_logger = logging.getLogger(__name__)

MIG_SERVICES = ("odoo", "oca")
MIG_OPERATIONS = (
    "upgrade",
    "install",
    "uninstall",
    "remove",
    "service",
    "pre_scripts",
    "post_scripts",
)

YAML_EXAMPLE = u"""
--- !Migration
version: 0.0.1
app_version: 10.0
upgrade:
- account
install:
- document
uninstall:
- project
remove:  # Only supported since Odoo 10.0
- removed_code

--- !Migration
version: 0.0.2  # nothing to do
app_version: 10.0

--- !Migration
version: 0.9.9  # prepare for migration service
app_version: 10.0
pre_scripts:
- ./migrate/my-pre-script.py
upgrade:
- document

--- !Migration
version: 1.0.0
app_version: 11.0
service: 'odoo'  # ['odoo'|'oca'] Migration provider
upgrade:  # Executed after migration service
- document
post_scripts:  # Executed after migration service
- ./migrate/my-post-script.py

"""


class Migration(yaml.YAMLObject):
    """ A single migration defined by a YAML document """

    __slots__ = ("version", "app_version") + MIG_OPERATIONS
    yaml_tag = u"!Migration"

    def __setstate__(self, data):
        """ yaml.load_all does call __setstate__, but not __init__ """
        self.__init__(**data)

    # Note: sets are not json serializable

    def __init__(
        self,
        version=None,
        app_version=None,
        upgrade=None,
        install=None,
        uninstall=None,
        remove=None,
        service=None,
        pre_scripts=None,
        post_scripts=None,
    ):
        self.version = semver.parse_version_info(version)
        self.app_version = app_version
        self.upgrade = self._validate_modules(upgrade, "upgrade")
        self.install = self._validate_modules(install, "install")
        self.uninstall = list(set(uninstall)) if uninstall else []
        self.remove = list(set(remove)) if remove else []
        self.service = self._validate_service(service)
        self.pre_scripts = list(set(pre_scripts)) if pre_scripts else []
        self.post_scripts = list(set(post_scripts)) if post_scripts else []

    @staticmethod
    def _validate_modules(obj, key):
        if obj is None:
            return []
        message = (
            "`{}` key accepts a list of modules that is present in " "addons paths."
        ).format(key)
        if not (obj and all(isinstance(elem, string_type) for elem in obj)):
            raise ParseError(message, YAML_EXAMPLE)
        # TODO: Check if modules are present in source code
        return list(set(obj))

    @staticmethod
    def _validate_service(obj):
        if obj is None:
            return None
        message = "`service` key accepts one of as string: {}".format(
            ", ".join(MIG_SERVICES)
        )
        if obj not in MIG_SERVICES:
            raise ParseError(message, YAML_EXAMPLE)
        return obj

    def _mark(self, env):
        """ Mark operations of this migration in the provided environment """
        with env.registry.cursor() as cursor:
            imm = env(cr=cursor)["ir.module.module"]
            for name in self.upgrade:
                imm.search([("name", "=", name)]).button_upgrade()
            for name in self.install:
                imm.search([("name", "=", name)]).button_install()
            for name in self.uninstall:
                imm.search([("name", "=", name)]).button_uninstall()

    def _remove(self, env):
        """ Cleanup module from ir.module.module
        The only use case is, if you completely drop a module without
        replacement. In case of a replacement, include this cleanup into the
        target's migration script or conditional init hoook."""
        if not migration:
            _logger.warning("remove operation is not supported for Odoo < 10.0.")
            return
        for name in self.remove:
            with env.registry.cursor() as cr:
                migration.remove_module(cr, name)

    def _run_pre_scripts(self, cr):
        for script in self.pre_scripts:
            exec(open(os.path.abspath(script)).read(), {"cr": cr})

    def _run_post_scripts(self, cr):
        for script in self.post_scripts:
            exec(open(os.path.abspath(script)).read(), {"cr": cr})

    def run(self, env):
        """ Run the actual migration """
        self._mark(env)
        env.reset()
        self._run_pre_scripts(env.cr)
        try:
            odoo.modules.registry.Registry.new(
                env.registry.db_name, update_module="migration"
            )
        except AttributeError:  # Odoo <= 9.0
            odoo.modules.registry.RegistryManager.new(
                env.registry.db_name, update_module="migration"
            )
        self._remove(env)
        self._run_post_scripts(env.cr)
        return env

    def is_noop(self):
        """ Check if Migration is a non operation """
        if not any(getattr(self, a, False) for a in MIG_OPERATIONS):
            return True
        return False


class MigrationSpec(object):
    """ A series of migrations loaded from a yaml file, bound to an
    environment. """

    def __init__(self, env, stream, since, until):
        self.migrations = sorted(
            {mig for mig in yaml.load_all(stream)}, key=lambda m: m.version
        )
        self.mig_table = MigrationTable(env)
        self.env = env
        self.since = since
        self.until = until

    def _is_applied(self, mig):
        return mig.version in self._get_finished_vers()

    def _get_finished_vers(self):
        return sorted(
            {
                semver.parse_version_info(v.number)
                for v in self.mig_table.versions()
                if v.date_done
            }
        )

    def _get_unfinished_vers(self):
        return sorted(
            {
                semver.parse_version_info(v.number)
                for v in self.mig_table.versions()
                if not v.date_done
            }
        )

    def _get_todo_migrations(self):
        if self.since and self.until:

            def _condition(mig):
                return mig.version > self.since and mig.version <= self.until

        elif self.since:

            def _condition(mig):
                return mig.version > self.since

        elif self.until:

            def _condition(mig):
                return mig.version <= self.until and not self._is_applied(mig)

        else:

            def _condition(mig):
                return not self._is_applied(mig)

        for mig in self.migrations:
            if _condition(mig):
                yield mig

    def run(self):
        """ Execute all applicable migrations from the spec """
        unfinished_v = self._get_unfinished_vers()
        if unfinished_v:
            strfmt = u",".join(unfinished_v)
            _logger.error("migrations %s are in unfinished state.", strfmt)
            raise MigrationErrorUnfinished(strfmt)

        if self.since and self.since not in self._get_finished_vers():
            _logger.error(
                "last migration %s not at par with %s.",
                self._get_finished_vers()[-1],
                self.since,
            )
            raise MigrationErrorGap(self._get_finished_vers()[-1], self.since)

        for mig in self._get_todo_migrations():
            # In case of --since dating to already applied verions
            if self._is_applied(mig):
                _logger.info(
                    u"migration %s is already applied - nothing to do.", (mig.version,)
                )
                continue

            self.mig_table.start(
                str(mig.version), mig.app_version, datetime.datetime.now()
            )

            if mig.is_noop():
                _logger.info(
                    u"migration %s is a non operation - only register bump.",
                    (mig.version,),
                )
            else:
                mig.run(self.env)

            self.mig_table.finish(
                str(mig.version),
                datetime.datetime.now(),
                {op: getattr(mig, op) for op in MIG_OPERATIONS},
            )



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)


EXOGENOUS_INFORMATION_SUGGESTION = (
    "If it's identifiable as part of a cross module relocation, modify the known changes file. "
    "Meaning: It has neither a processable oldname instruction, nor is it a relocatino alongside "
    "the inheritence MRO, which are matched automatically."
)
TABLE_RENAME_DEFINITION = 'A relocation of all non-automatic columns of a table is also known as a "table rename".'
TRANSITIONS_LEVELS = {
    "0": "Nothing to do.",
    "1": "Automateable transition.",
    "2": "Exogenous transition.",
}


# fmt: off
KNOWN_SCHEMA_ATOMIC_TRANSITIONS_DESCRIPTIONS = {
    'gained_existence'                      : {'description': "Column will appear in module: {module}."},  # {module} first definition as per MRO, apply oldname + EXOGENOUS_INFORMATION preflight check
    'lost_existence'                        : {'description': "Column will disappear in module: {module}."},  # {module} first definition as per MRO, apply oldname + EXOGENOUS_INFORMATION preflight check
    'gained__modules'                       : {'description': "Column will newly be defined in: {_modules}"},
    'lost__modules'                         : {'description': "Column won't be defined any more in: {_modules}"},

    'gained_required'                       : {'description': "Column will become non nullable."},
    'lost_required'                         : {'description': "Column will become nullable."},
    'gained_type'                           : {'description': "Column will be of type {type}."},
    'lost_type'                             : {'description': "Column won't be of type {type} any more."},

    # Same type transitions (type transition shadow attriute transitions)

    # Mutually exlusive store transitions and it's variants ordered from complex to trivial
    'gained_compute'                        : {'description': "Column will be computed."},
    'lost_cmopute_stored'                   : {'description': "Column won't be computed any more, but was stored."},
    'lost_cmopute_not_stored'               : {'description': "Column won't be computed any more and wasn't stored."},
    'gained_related'                        : {'description': "Column will be related to {related}."},
    'lost_related_stored'                   : {'description': "Column won't be related to {related} any more, but was stored."},
    'lost_related_not_stored'               : {'description': "Column won't be related to {related} any more and wasn't stored."},
    'gained_company_dependent'              : {'description': "Column will be company dependent."},
    'lost_company_dependent'                : {'description': "Column won't be company dependent any more."},  # Note: company_dependent are never stored
    'gained_store'                          : {'description': "Column will be persisted in database."},
    'lost_store'                            : {'description': "Column won't be persisted in database any more."},

    'gained_deprecated'                     : {'description': "Column will be deprecated."},
    'lost_deprecated'                       : {'description': "Column won't be deprecated any more."},  # probably never gonna happen
    'gained_size'                           : {'description': "Column will be of size {size}."},
    'lost_size'                             : {'description': "Column won't be of size {size} any more."},
    'gained_attachment'                     : {'description': "Column will be stored as attachment."},
    'lost_attachment'                       : {'description': "Column won't be stored as attachment any more."},
    'gained_selection'                      : {'description': "Column will have selection {selection}."},
    'lost_selection'                        : {'description': "Column won't have selection {selection} any more."},
}


KNOWN_SCHEMA_AGGREGATE_TRANSITIONS_DESCRIPTIONS = {
    'moved_oldname'                         : {'description': "Column will be moved / renamed / relocated according to `oldname` field."},
    'moved_mro_parallel'                    : {'description': "Column will be moved / renamed / relocated along the inheritence MRO."},
    'moved_exogenous'                       : {'description': "Column will be moved / renamed / relocated according to exogenous information."},
    # Casted type transitions
    'casted'                                : {'description': "Column will be automatically type casted by the system."},

    # None-Casted Type transitions
    'from_selection_to_many2one'            : {'description': "Column will change from `selection` to `many2one`."},
    'from_many2one_to_selection'            : {'description': "Column will change from `many2one` to `selection`."},
    'from_selection_to_boolean'             : {'description': "Column will change from `selection` to `boolean`."},
    'from_boolean_to_selection'             : {'description': "Column will change from `boolean` to `selection`."},
    'from_selection_to_char'                : {'description': "Column will change from `selection` to `char`."},
    'from_char_to_selection_casted'         : {'description': "Column will change from `char` to `selection` (casted)."},
    'from_char_to_selection_not_casted'     : {'description': "Column will change from `char` to `selection` (not casted)."},
    'from_char_to_selection_invalid'        : {'description': "Column will change from `char` to `selection` (casted & invalid)."},
    'from_selection_to_integer'             : {'description': "Column will change from `selection` to `integer`."},
    'from_integer_to_selection_casted'      : {'description': "Column will change from `integer` to `selection` (casted)."},
    'from_integer_to_selection_not_casted'  : {'description': "Column will change from `integer` to `selection` (not casted)."},
    'from_integer_to_selection_invalid'     : {'description': "Column will change from `integer` to `selection` (casted & invalid)."},

    'from_many2many_to_many2one'            : {'description': "Column will change from `many2many` to `many2one`."},
    'from_many2one_to_many2many'            : {'description': "Column will change from `many2one` to `many2many`."},

    'from_float_to_integer'                 : {'description': "Column will change from `float` to `integer`."},
    'from_monetary_to_integer'              : {'description': "Column will change from `monetary` to `integer`."},  # Really?
    'from_binary_to_text_or_char_or_html'   : {'description': "Column will change from `binary` to `text` or `char` or `html`."},
    'from_text_or_char_or_html_to_binary'   : {'description': "Column will change from `text` or `char` or `html` to `binary`."},

    'not_known_not_casted'                  : {'description': "Column transition not known."},

    # Same type transitions
    'char_shrinks'                          : {'description': "Column (char type) will shrink in size."},
    'char_grows'                            : {'description': "Column (char type) will grow in size."},
    'selection_shrinks'                     : {'description': "Column (selection type) will have less options, missing options: {options}."},
    'selection_grows'                       : {'description': "Column (selection type) will have more options, additional options: {options}."},
}

# migration_semantic only defines `pre` or `post` (no `end`)
# pre are meant to be casted to the earliest possible module
# post are meant to be casted to the latest possible module
# Hence `end` semantic is a disguised `post` semantic

KNOWN_SCHEMA_ATOMIC_TRANSITIONS_INSTRUCTIONS = {
    'gained_existence'                      : {'instruction': "Pre migration can seed column values. " + EXOGENOUS_INFORMATION_SUGGESTION,                          "migration_semantic": 'pre', 'level': 2},
    'lost_existence'                        : {'instruction': "Nothing to do. It will be gone. " + EXOGENOUS_INFORMATION_SUGGESTION,                                "migration_semantic": None,  'level': 0},
    'gained__modules'                       : {'instruction': "Informational, together with module dependencies, for proper migration script (stub) casting.",      "migration_semantic": None,  'level': 2},  # Always report
    'lost__modules'                         : {'instruction': "Informational, together with module dependencies, for proper migration script (stub) casting.",      "migration_semantic": None,  'level': 2},  # Always report

    'gained_required'                       : {'instruction': "Pre migration needs to manually induce correct semantics for NULL values in that column.",           "migration_semantic": 'pre', 'level': 2},
    'lost_required'                         : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},
    'gained_type'                           : {'instruction': "Nothing to do. It's part of an aggregate transition.",                                               "migration_semantic": None,  'level': 0},
    'lost_type'                             : {'instruction': "Nothing to do. It's part of an aggregate transition.",                                               "migration_semantic": None,  'level': 0},

    # Same type transitions (type transition shadow attriute transitions)

    # Mutually exlusive store transitions and it's variants ordered from complex to trivial
    'gained_compute'                        : {'instruction': "Pre migration will need to shift (+merge) data to the sources used by the new compute function.",    "migration_semantic": 'pre', 'level': 2},
    'lost_compute_stored'                   : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},
    'lost_compute_not_stored'               : {'instruction': "Pre migration will need to force storage and trigger a recompute on old codebase.",                  "migration_semantic": 'pre', 'level': 1},
    'gained_related'                        : {'instruction': "Pre migration might need to shift (+merge) data to related target.",                                 "migration_semantic": 'pre', 'level': 1},
    'lost_related_stored'                   : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},
    'lost_related_not_stored'               : {'instruction': "Pre migration will need to force storage and trigger a recompute on old codebase.",                  "migration_semantic": 'pre', 'level': 1},
    'gained_company_dependent'              : {'instruction': "Pre migration will need to shift data to ir_property table once for every company.",                 "migration_semantic": 'pre', 'level': 1},
    'lost_company_dependent'                : {'instruction': "Pre migration will need to homologate diverging companies' values and set the chosen value.",        "migration_semantic": 'pre', 'level': lambda env: 1 if len(env['res.compay'].search() == 1) else 2},  # TODO: In reality it should target cardinality of property values
    'gained_store'                          : {'instruction': "Nothing to do. It's most probably already a computed type field.",                                   "migration_semantic": None,  'level': 0},  # Note: Never seen a non computed type with store=false
    'lost_store'                            : {'instruction': "Nothing to do. If so data will be shifted by a more specific transition",                            "migration_semantic": None,  'level': 0},

    'gained_deprecated'                     : {'instruction': "Pre migration might want to get away from this field.",                                              "migration_semantic": 'pre', 'level': 2},
    'lost_deprecated'                       : {'instruction': "Maybe it's a joke, or 1st of april. If you see this message, then open an issue somwhere.",          "migration_semantic": None,  'level': 0},
    'gained_size'                           : {'instruction': "Nothing to do. It's part of an aggregate transition.",                                               "migration_semantic": None,  'level': 0},
    'lost_size'                             : {'instruction': "Nothing to do. It's part of an aggregate transition.",                                               "migration_semantic": None,  'level': 0},
    'gained_attachment'                     : {'instruction': "Pre migration needs to transfer byte data from the column to the attachment table.",                 "migration_semantic": 'pre', 'level': 1},
    'lost_attachment'                       : {'instruction': "Pre migration needs to transfer byte data from the attachment table to the column.",                 "migration_semantic": 'pre', 'level': 1},
    'gained_selection'                      : {'instruction': "Nothing to do. It's part of an aggregate transition.",                                               "migration_semantic": None,  'level': 0},
    'lost_selection'                        : {'instruction': "Nothing to do. It's part of an aggregate transition.",                                               "migration_semantic": None,  'level': 0},
}

KNOWN_SCHEMA_AGGREGATE_TRANSITIONS_INSTRUCTIONS = {
    'moved_oldname'                         : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},
    'moved_mro_parallel'                    : {'instruction': "Pre migration will need to relocate the column.",                                                    "migration_semantic": 'pre', 'level': 1},
    'moved_exogenous'                       : {'instruction': "Pre migration will need to relocate the column. " + TABLE_RENAME_DEFINITION,                         "migration_semantic": 'pre', 'level': 1},
    # Casted type transitions
    'casted'                                : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},

    # None-Casted Type transitions
    'from_selection_to_many2one'            : {'instruction': "Pre migration will need to map selection char or int to foreign ids.",                               "migration_semantic": 'pre', 'level': 2},
    'from_many2one_to_selection'            : {'instruction': "Pre migration will need to map foreign ids to selecton char or int.",                                "migration_semantic": 'pre', 'level': 2},
    'from_selection_to_boolean'             : {'instruction': "Pre migration will need to map selection char or into to boolean values.",                           "migration_semantic": 'pre', 'level': 2},
    'from_boolean_to_selection'             : {'instruction': "Pre migration will need to map boolean values to selecton char or int.",                             "migration_semantic": 'pre', 'level': 2},
    'from_selection_to_char'                : {'instruction': "Pre migration will need to map selection int to char values (not casted).",                          "migration_semantic": 'pre', 'level': 2},
    'from_char_to_selection_casted'         : {'instruction': "Nothing to do. Sometimes casted selections can validate unintendetly: double check final result.",   "migration_semantic": None,  'level': 0},
    'from_char_to_selection_not_casted'     : {'instruction': "Pre migration will need to map char values to selection int (not casted).",                          "migration_semantic": 'pre', 'level': 2},
    'from_char_to_selection_invalid'        : {'instruction': "Pre migration will need to map int values to selection int (invalid).",                              "migration_semantic": 'pre', 'level': 2},
    'from_selection_to_integer'             : {'instruction': "Nothing to do. Sometimes casted selections can validate unintendetly: double check final result.",   "migration_semantic": None,  'level': 0},
    'from_integer_to_selection_casted'      : {'instruction': "Pre migration will need to map int values to selection char (not casted).",                          "migration_semantic": 'pre', 'level': 2},
    'from_integer_to_selection_not_casted'  : {'instruction': "Pre migration will need to map int values to selection char (not casted).",                          "migration_semantic": 'pre', 'level': 2},
    'from_integer_to_selection_invalid'     : {'instruction': "Pre migration will need to map char values to selection char (invalid).",                            "migration_semantic": 'pre', 'level': 2},

    'from_many2many_to_many2one'            : {'instruction': "Pre migration will need to homologate a single relational value into a new column.",                 "migration_semantic": 'pre', 'level': 2},
    'from_many2one_to_many2many'            : {'instruction': "Pre migration might need to shift single reference into rel table.",                                 "migration_semantic": 'pre', 'level': 1},

    'from_float_to_integer'                 : {'instruction': "Pre migration will need to round values.",                                                           "migration_semantic": 'pre', 'level': 1},
    'from_monetary_to_integer'              : {'instruction': "Pre migration will need to round values. (Note: it's suspicious a monetary field becomes integer)",  "migration_semantic": 'pre', 'level': 1},
    'from_binary_to_text_or_char_or_html'   : {'instruction': "Pre migration will need to decode binary data into string.",                                         "migration_semantic": 'pre', 'level': 1},
    'from_text_or_char_or_html_to_binary'   : {'instruction': "Pre migration will need to encode string into binary data.",                                         "migration_semantic": 'pre', 'level': 1},

    'not_known_not_casted'                  : {'instruction': "Pre migration needs to transform data in unconventional ways.",                                      "migration_semantic": 'pre', 'level': 2},

    # Same type transitions
    'char_shrinks'                          : {'instruction': "Pre migration will need to identify and trunkate oversized chars.",                                  "migration_semantic": 'pre', 'level': 2},
    'char_grows'                            : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},
    'selection_shrinks'                     : {'instruction': "Pre migration will need to homologate off values to existing ones.",                                 "migration_semantic": 'pre', 'level': 2},
    'selection_grows'                       : {'instruction': "Nothing to do.",                                                                                     "migration_semantic": None,  'level': 0},
}

KNOWN_SCHEMA_ATOMIC_TRANSITIONS = KNOWN_SCHEMA_ATOMIC_TRANSITIONS_DESCRIPTIONS
KNOWN_SCHEMA_ATOMIC_TRANSITIONS.update(KNOWN_SCHEMA_ATOMIC_TRANSITIONS_INSTRUCTIONS)
KNOWN_SCHEMA_AGGREGATE_TRANSITIONS = KNOWN_SCHEMA_AGGREGATE_TRANSITIONS_DESCRIPTIONS
KNOWN_SCHEMA_AGGREGATE_TRANSITIONS.update(KNOWN_SCHEMA_AGGREGATE_TRANSITIONS_INSTRUCTIONS)
# fmt: on



# -*- coding: utf-8 -*-
# Copyright 2016-2017 Camptocamp SA
# Copyright 2017-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import json
from collections import namedtuple


class MigrationTable(object):
    def __init__(self, env):
        self.env = env
        self.table_name = "dodoo_migrator"
        self.VersionRecord = namedtuple(
            "VersionRecord", "number app_version date_start date_done operations"
        )
        self._versions = None
        self._create_if_not_exists()

    def _create_if_not_exists(self):
        with self.env.registry.cursor() as cursor:
            query = """
            CREATE TABLE IF NOT EXISTS {} (
                number VARCHAR NOT NULL,
                app_version VARCHAR NOT NULL,
                date_start TIMESTAMP NOT NULL,
                date_done TIMESTAMP,
                operations TEXT,

                CONSTRAINT version_pk PRIMARY KEY (number)
            );
            """.format(
                self.table_name
            )
            cursor.execute(query)

    def versions(self):
        """ Read versions from the table

        The versions are kept in cache for the next reads.
        """
        if self._versions is not None:
            return self._versions
        with self.env.registry.cursor() as cursor:
            query = """
            SELECT number,
                   app_version,
                   date_start,
                   date_done,
                   operations
            FROM {}
            """.format(
                self.table_name
            )
            cursor.execute(query)
            rows = cursor.fetchall()
            versions = []
            for row in rows:
                row = list(row)
                # convert 'operations' to json
                row[4] = json.loads(row[4]) if row[4] else []
                versions.append(self.VersionRecord(*row))
            self._versions = versions
        return self._versions

    def start(self, version, app_version, timestamp):
        with self.env.registry.cursor() as cursor:
            query = """
            INSERT INTO {}
            (number, app_version, date_start)
            VALUES (%s, %s, %s)
            """.format(
                self.table_name
            )
            cursor.execute(query, (version, app_version, timestamp))
        self._versions = None  # reset versions cache

    def finish(self, version, timestamp, operations):
        with self.env.registry.cursor() as cursor:
            query = """
            UPDATE {}
            SET date_done = %s,
                operations = %s
            WHERE number = %s
            """.format(
                self.table_name
            )
            cursor.execute(query, (timestamp, json.dumps(operations), version))
            self._versions = None  # reset versions cache



# -*- coding: utf-8 -*-
# Copyright 2016-2017 Camptocamp SA
# Copyright 2017-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import click as _click


class _MigrationError(Exception):
    pass


class MigrationErrorUnfinished(_MigrationError):
    def __init__(self, unfinished):
        super(MigrationErrorUnfinished, self).__init__(unfinished)
        self.unfinished = unfinished

    def __str__(self):
        msg = (
            u"Upgrade of version {} has been attempted and failed. "
            u"You may want to restore the backup or to run again the "
            u"migration with the --force flag or to fix it manually "
            u"In that case, you will have to "
            u"update the  'dodoo_migrator' table yourself.".format(self.unfinished)
        )
        return msg


class MigrationErrorGap(_MigrationError):
    def __init__(self, finished, since):
        super(MigrationErrorGap, self).__init__(finished, since)
        self.finished = finished
        self.since = since

    def __str__(self):
        msg = (
            u"The --since flag must specify the last executed migration. "
            u"Specifying a later migration would cause a igration gap. "
            u"Last finished version {}. "
            u"--since {}. ".format(self.finished, self.since)
        )
        return msg


class ParseError(_click.ClickException):
    def __init__(self, message, example=None):
        super(ParseError, self).__init__(message)
        self.example = example

    def __str__(self):
        if not self.example:
            return super(ParseError, self).__str__()
        msg = (
            u"An error occured during the parsing of the configuration "
            u"file. Here is an example to help you to figure out "
            u"your issue.\n{}\n{}"
        ).format(self.example, self.args[0])
        return msg



def migrate(cr, version):
    query = """
    CREATE TABLE IF NOT EXISTS dodoo_test_migrations (
        name VARCHAR NOT NULL
    );
    """
    cr.execute(query)

    cr.execute("""INSERT INTO dodoo_test_migrations VALUES ('pre')""")



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import subprocess

import click


class Git(object):
    def __enter__(self):
        self.head = self.get_branch_name()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.checkout(self.head)

    def __init__(self, git_dir):
        self.git_dir = git_dir
        self.head = None
        click.echo("==> Git-Dir: %s" % self.git_dir)

    def run(self, command):
        """Execute git command in bash
        :param list command: Git cmd to execute in self.git_dir
        :return: String output of command executed.
        """
        cmd = ["git", "--git-dir=" + self.git_dir] + command
        try:
            click.echo(">>> " + " ".join(cmd))
            res = subprocess.check_output(cmd)
        except subprocess.CalledProcessError:
            res = None
        if isinstance(res, bytes):
            res = res.decode("utf-8")
        if res:
            res = res.strip("\n")
        return res

    def checkout(self, branch):
        command = ["checkout", "--recurse-submodules", branch]
        res = self.run(command)
        if res is None:
            ctx = click.get_current_context()
            ctx.fail("Checkout failed. Aborting for security.")

    def get_branch_name(self):
        """Get branch name
        :return: String with name of current branch name"""
        command = ["rev-parse", "--abbrev-ref", "HEAD"]
        return self.run(command)



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)


#
# THIS FILE IS A 1:1 RE IMPLEMENTATION OF THE ODOO MIGRATION MANAGER WITH ONLY
# ONE DIFFERENCE: Allow to *lay over* a migration folder layout.
#

from __future__ import absolute_import, print_function

import glob
import logging
import os
import sys

from dodoo import odoo

from .cli import get_additional_mig_path

# We need to adopt this strange pattern, as in p27 the import resolution would
# be fooled by the src.odoo package, meant to blend in with the odoo namespace
# from odoo import ... would not use real odoo, but src.odoo in py27
MigrationManager = odoo.modules.migration.MigrationManager  # noqa
parse_version = odoo.tools.parse_version  # noqa

if sys.version_info[0] == 2:
    import imp

    def load_script(path, module_name):
        fp = open(path, "r")
        fname = path
        fp2 = None

        # pylint: disable=file-builtin,undefined-variable
        if not isinstance(fp, file):  # noqa
            # imp.load_source need a real file object, so we create
            # one from the file-like object we get from file_open
            fp2 = os.tmpfile()
            fp2.write(fp.read())
            fp2.seek(0)

        try:
            return imp.load_source(module_name, fname, fp2 or fp)
        finally:
            if fp:
                fp.close()
            if fp2:
                fp2.close()


else:
    import importlib.util

    def load_script(path, module_name):
        spec = importlib.util.spec_from_file_location(module_name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module


try:
    get_resource_path = odoo.modules.module.get_resource_path  # noqa
except AttributeError:  # Odoo < 9.0

    get_module_path = odoo.modules.module.get_module_path  # noqa

    def get_resource_path(module, *args):
        """Return the full path of a resource of the given module.
        :param module: module name
        :param list(str) args: resource path components within module
        :rtype: str
        :return: absolute path to the resource
        TODO make it available inside on osv object (self.get_resource_path)
        """
        mod_path = get_module_path(module)
        if not mod_path:
            return False
        resource_path = os.path.join(mod_path, *args)
        if os.path.isdir(mod_path):
            # the module is a directory - ignore zip behavior
            if os.path.exists(resource_path):
                return resource_path
        return False


_logger = logging.getLogger(__name__)


def _get_additional_migration_path(module, *args):
    if not get_additional_mig_path():
        return False
    mod_path = os.path.join(get_additional_mig_path(), module)
    if not mod_path:
        return False
    resource_path = os.path.join(mod_path, *args)
    if os.path.isdir(mod_path):
        # the module is a directory - ignore zip behavior
        if os.path.exists(resource_path):
            return resource_path
    return False


class ExtendedMigrationManager(MigrationManager):
    def _get_files(self):
        def _get_scripts(res, path):
            for version in os.listdir(path):
                if not os.path.isdir(os.path.join(path, version)):
                    continue
                files = glob.glob1(os.path.join(path, version), "*.py")
                files.sort()
                if version not in res:
                    res[version] = [
                        os.path.join(path, version) + os.path.sep + f for f in files
                    ]
                else:
                    res[version].append(
                        [os.path.join(path, version) + os.path.sep + f for f in files]
                    )

        def get_scripts(default, overlay):
            res = {}
            if default:
                _get_scripts(res, default)
            if overlay:
                _get_scripts(res, overlay)
            return res

        for pkg in self.graph:
            if not (
                hasattr(pkg, "update")
                or pkg.state == "to upgrade"  # noqa: W504
                or getattr(pkg, "load_state", None) == "to upgrade"  # noqa: W504
            ):
                continue

            module = get_scripts(
                get_resource_path(pkg.name, "migrations"),
                _get_additional_migration_path(pkg.name, "migrations"),
            )
            maintenance = get_scripts(
                get_resource_path("base", "maintenance", "migrations", pkg.name),
                _get_additional_migration_path(
                    "base", "maintenance", "migrations", pkg.name
                ),
            )

            self.migrations[pkg.name] = {"module": module, "maintenance": maintenance}

    def migrate_module(self, pkg, stage):
        assert stage in ("pre", "post", "end")
        stageformat = {"pre": "[>%s]", "post": "[%s>]", "end": "[$%s]"}
        state = (
            pkg.state if stage in ("pre", "post") else getattr(pkg, "load_state", None)
        )
        if (
            not (hasattr(pkg, "update") or state == "to upgrade")
            or state == "to install"
        ):
            return

        def _convert_version(version):
            if version.count(".") >= 2:
                return version  # the version number already containt the server version
            return "{}.{}".format(odoo.release.major_version, version)

        def _get_migration_versions(pkg):
            versions = sorted(
                {
                    ver
                    for lv in self.migrations[pkg.name].values()
                    for ver, lf in lv.items()
                    if lf
                },
                key=lambda k: parse_version(_convert_version(k)),
            )
            return versions

        def _get_migration_files(pkg, version, stage):
            """ return a list of migration script files
            """
            migrations = self.migrations[pkg.name]
            lst = []

            for _, values in migrations.items():
                if version in values:
                    for file in values[version]:
                        if not os.path.basename(file).startswith(stage + "-"):
                            continue
                        lst.append(file)
            lst.sort()
            return lst

        installed_version = getattr(pkg, "load_version", pkg.installed_version) or ""
        parsed_installed_version = parse_version(installed_version)
        current_version = parse_version(_convert_version(pkg.data["version"]))

        versions = _get_migration_versions(pkg)

        for version in versions:
            if (
                parsed_installed_version
                < parse_version(_convert_version(version))
                <= current_version
            ):

                strfmt = {
                    "addon": pkg.name,
                    "stage": stage,
                    "version": stageformat[stage] % version,
                }

                for pyfile in _get_migration_files(pkg, version, stage):
                    name, ext = os.path.splitext(os.path.basename(pyfile))
                    if ext.lower() != ".py":
                        continue
                    mod = None
                    try:
                        mod = load_script(pyfile, name)
                        _logger.info(
                            "module %(addon)s: Running migration"
                            " %(version)s %(name)s" % dict(strfmt, name=mod.__name__)
                        )
                        migrate = mod.migrate
                    except ImportError:
                        _logger.exception(
                            "module %(addon)s: Unable to load"
                            "%(stage)s-migration file"
                            "%(file)s" % dict(strfmt, file=pyfile)
                        )
                        raise
                    except AttributeError:
                        _logger.error(
                            "module %(addon)s: Each %(stage)s-"
                            'migration file must have a "migrate(cr,'
                            ' installed_version)" function' % strfmt
                        )
                    else:
                        migrate(self.cr, installed_version)
                    finally:
                        if mod:
                            del mod


odoo.modules.migration.MigrationManager = ExtendedMigrationManager



# -*- coding: utf-8 -*-
# Part of Odoo. See LICENSE file for full copyright and licensing details.

import datetime
import imp
import json
import logging
import os
import re
import sys
import time
from collections import namedtuple
from contextlib import contextmanager
from functools import reduce
from inspect import currentframe
from itertools import chain, islice
from operator import itemgetter
from textwrap import dedent

import lxml
import markdown
import psycopg2
from docutils.core import publish_string
from dodoo import odoo
from psycopg2 import sql

from odoo import SUPERUSER_ID, release
from odoo.api import Environment
from odoo.modules.module import get_module_path
from odoo.modules.registry import Registry
from odoo.sql_db import db_connect
from odoo.tools import UnquoteEvalContext
from odoo.tools.func import frame_codeinfo
from odoo.tools.mail import html_sanitize

try:
    from odoo.addons.base.models.ir_module import MyWriter
except ImportError:
    # Odoo < 12.0
    from odoo.addons.base.module.module import MyWriter


_logger = logging.getLogger(__name__)

_INSTALLED_MODULE_STATES = ("installed", "to install", "to upgrade")
IMD_FIELD_PATTERN = "field_%s__%s"

DROP_DEPRECATED_CUSTOM = os.getenv("OE_DROP_DEPRECATED_CUSTOM")

# migration environ, used to share data between scripts
ENVIRON = {}


class MigrationError(Exception):
    pass


# Modules utilities


def modules_installed(cr, *modules):
    """Check if the provided modules are installed (or about to be)."""
    assert modules
    _query = sql.SQL(
        """

        SELECT count(1)
        FROM ir_module_module
        WHERE name IN %(names)s
          AND state IN %(states)s

    """
    )
    cr.execute(_query, dict(names=modules, states=_INSTALLED_MODULE_STATES))
    return cr.fetchone()[0] == len(modules)


def module_installed(cr, module):
    """Check if the provided module is installed (or about to be)."""
    return modules_installed(cr, module)


def remove_module(cr, module):
    """Uninstall the module and delete all its xmlid entries.

    Make sure to reassign records before calling this method.
    """
    # NOTE: we cannot use the uninstall of module because the given
    # module need to be currenctly installed and running as deletions
    # are made using orm.

    _query = sql.SQL(
        """

        SELECT id
        FROM ir_module_module WHERE name = %(module)s

    """
    )
    cr.execute(_query, locals())
    mod_id, = cr.fetchone() or [None]
    if not mod_id:
        return

    # delete constraints only owned by this module
    _query = sql.SQL(
        """

        SELECT name
        FROM ir_model_constraint
        GROUP BY name HAVING array_agg(module) = %(mod_ids)s

    """
    )
    cr.execute(_query, {"mod_ids": [mod_id]})

    constraints = tuple(vals[0] for vals in cr.fetchall())
    if constraints:
        _query = sql.SQL(
            """

            SELECT TABLE_NAME,
                   CONSTRAINT_NAME
            FROM information_schema.table_constraints
            WHERE CONSTRAINT_NAME IN %(constraints)s

        """
        )
        cr.execute(_query, locals())

        for table, constraint in cr.fetchall():
            _params = {
                "table": sql.Identifier(table),
                "constraint": sql.Identifier(constraint),
            }
            _query = sql.SQL(
                """

                ALTER TABLE {table}
                DROP CONSTRAINT {constraint}

            """
            ).format(**_params)
            cr.execute(_query)

    _query = sql.SQL(
        """

        DELETE
        FROM ir_model_constraint
        WHERE module = %(mod_id)s

    """
    )
    cr.execute(_query, locals())

    # delete data
    model_ids, field_ids, view_ids, menu_ids = (), (), (), ()
    _query = sql.SQL(
        """

        SELECT model,
               array_agg(res_id)
        FROM ir_model_data d
        WHERE NOT EXISTS
            (SELECT 1
             FROM ir_model_data
             WHERE id != d.id
               AND res_id = d.res_id
               AND model = d.model
               AND module != d.module)
          AND module = %(module)s
          AND model != 'ir.module.module'
        GROUP BY model

    """
    )
    cr.execute(_query, locals())
    for model, res_ids in cr.fetchall():
        if model == "ir.model":
            model_ids = tuple(res_ids)
        elif model == "ir.model.fields":
            field_ids = tuple(res_ids)
        elif model == "ir.ui.view":
            view_ids = tuple(res_ids)
        elif model == "ir.ui.menu":
            menu_ids = tuple(res_ids)
        else:
            table = table_of_model(cr, model)
            if table_exists(cr, table):
                _params = {"table": sql.Identifier(table)}
                _query = sql.SQL(
                    """

                    DELETE
                    FROM {table}
                    WHERE id IN %(res_ids)s

                """
                ).format(**_params)
                cr.execute(_query, dict(res_ids=tuple(res_ids)))

    for view_id in view_ids:
        remove_view(cr, view_id=view_id, deactivate_custom=True, silent=True)

    if menu_ids:
        remove_menus(cr, menu_ids)

    # remove relations
    _query = sql.SQL(
        """

        SELECT name
        FROM ir_model_relation
        GROUP BY name HAVING array_agg(module) = %(mod_ids)s

    """
    )
    cr.execute(_query, {"mod_ids": [mod_id]})

    relations = tuple(vals[0] for vals in cr.fetchall())

    _query = sql.SQL(
        """

        DELETE
        FROM ir_model_relation
        WHERE module = %(mod_id)s

    """
    )
    cr.execute(_query, locals())
    if relations:
        _query = sql.SQL(
            """

            SELECT TABLE_NAME
            FROM information_schema.tables
            WHERE TABLE_NAME IN %(relations)s

        """
        )
        cr.execute(_query, locals())
        # fmt: on
        for (rel,) in cr.fetchall():
            _params = {"rel": sql.Identifier(rel)}
            _query = sql.SQL(
                """

            DROP TABLE {rel} CASCADE

            """
            ).format(**_params)
            cr.execute(_query)

    if model_ids:
        _query = sql.SQL(
            """

            SELECT model
            FROM ir_model
            WHERE id IN %(model_ids)s

        """
        )
        cr.execute(_query, locals())
        for (model,) in cr.fetchall():
            remove_model(cr, model)

    if field_ids:
        _query = sql.SQL(
            """

            SELECT model,
                   name
            FROM ir_model_fields
            WHERE id IN %(field_ids)s

        """
        )
        cr.execute(_query, locals())
        for model, name in cr.fetchall():
            remove_field(cr, model, name)

    _query = sql.SQL(
        """

        DELETE
        FROM ir_model_data
        WHERE model = 'ir.module.module'
          AND res_id = %(mod_id)s;


        DELETE
        FROM ir_model_data
        WHERE module = %(module)s;


        DELETE
        FROM ir_module_module
        WHERE name = %(module)s;


        DELETE
        FROM ir_module_module_dependency
        WHERE name = %(module)s;

    """
    )
    cr.execute(_query, locals())


def rename_module(cr, old, new):
    """Rename a module. Yes, really."""

    mod_old = "module_" + old
    mod_new = "module_" + new

    _query = sql.SQL(
        """

        UPDATE ir_module_module
        SET name = %(new)s
        WHERE name = %(old)s;


        UPDATE ir_module_module_dependency
        SET name = %(new)s
        WHERE name = %(old)s;


        UPDATE ir_model_data
        SET module = %(new)s
        WHERE module = %(old)s;


        UPDATE ir_model_data
        SET name = %(mod_new)s
        WHERE name = %(mod_old)s
          AND module = 'base'
          AND model = 'ir.module.module';

    """
    )
    cr.execute(_query, locals())


def merge_module(cr, old, into, tolerant=False):
    """Move all references of module `old` into module `into`.

        :param str old: source module to merge
        :param str into: destination module to merge into
        :tolerant bool: if True, the merge will be skipped if the database
                        does not have the `old` module in its ir_module_module
                        table (e.g. if the module was released after the
                        creation of the database)
    """

    _query = sql.SQL(
        """

        SELECT name,
               id
        FROM ir_module_module
        WHERE name IN %(names)s

    """
    )
    cr.execute(_query, dict(names=(old, into)))
    mod_ids = dict(cr.fetchall())

    if tolerant and old not in mod_ids:
        # this can happen in case of temp modules added after a release if the database does not
        # know about this module, i.e: account_full_reconcile in 9.0
        # `into` should be known; let it crash if not
        _logger.warning("Unknow module %s. Skip merging into %s.", old, into)
        return

    def _update(table, old, new):
        _query = sql.SQL(
            """

            UPDATE ir_model_{table} x
            SET module = %(new)s
            WHERE module = %(old)s
              AND NOT EXISTS
                (SELECT 1
                 FROM ir_model_{table} y
                 WHERE y.name = x.name
                   AND y.module = %(new)s)

        """.format(
                table=table
            )
        )
        cr.execute(_query, locals())

        if table == "data":
            _query = sql.SQL(
                """

                SELECT model,
                       array_agg(res_id)
                FROM ir_model_data
                WHERE module = %(old)s
                  AND model NOT LIKE 'ir.model%%'
                GROUP BY model

            """
            )
            cr.execute(_query, locals())
            for model, res_ids in cr.fetchall():
                # we can assume other records have been moved to xml files of the new module
                # remove the unnecessary data and let the module update do its job
                if model == "ir.ui.view":
                    for v in res_ids:
                        remove_view(cr, view_id=v, deactivate_custom=True, silent=True)
                elif model == "ir.ui.menu":
                    remove_menus(cr, tuple(res_ids))
                else:
                    for r in res_ids:
                        remove_record(cr, (model, r))

        _query = sql.SQL(
            """

            DELETE
            FROM ir_model_{table}
            WHERE module = %(old)s

        """.format(
                table=table
            )
        )
        cr.execute(_query, locals())

    _update("constraint", mod_ids[old], mod_ids[into])
    _update("relation", mod_ids[old], mod_ids[into])
    _update_view_key(cr, old, into)
    _update("data", old, into)

    # update dependencies

    _query = sql.SQL(
        """


        INSERT INTO ir_module_module_dependency(module_id, name)
        SELECT module_id,
               %(into)s
        FROM ir_module_module_dependency d
        WHERE name = %(old)s
          AND NOT EXISTS
            (SELECT 1
             FROM ir_module_module_dependency o
             WHERE o.module_id = d.module_id
               AND o.name = %(into)s);


        DELETE
        FROM ir_module_module
        WHERE name = %(old)s;


        DELETE
        FROM ir_module_module_dependency
        WHERE name = %(old)s;

    """
    )
    cr.execute(_query, locals())


def force_install_module(cr, module, deps=None):
    """Force a module to be installed during the upgrade process.

        :param str module: technical name of the module to install
        :param list deps: if set, the module will be installed only if any of
                          module of this list is already installed
                          (or about to be)
    """
    subquery = sql.SQL("")
    if deps:
        subquery = sql.SQL(
            """
            AND EXISTS
              (SELECT 1
               FROM ir_module_module
               WHERE name IN %(deps)s
                 AND state IN %(states)s)

        """
        )
        deps = tuple(deps)
        states = _INSTALLED_MODULE_STATES

    _params = {"subquery": subquery}
    _query = sql.SQL(
        """

         WITH RECURSIVE deps (mod_id, dep_name) AS
          (SELECT m.id,
                  d.name
           FROM ir_module_module_dependency d
           JOIN ir_module_module m ON (d.module_id = m.id)
           WHERE m.name = %(module)s
           UNION SELECT m.id,
                        d.name
           FROM ir_module_module m
           JOIN deps ON deps.dep_name = m.name
           JOIN ir_module_module_dependency d ON (d.module_id = m.id))
        UPDATE ir_module_module m
        SET state = CASE
                        WHEN state = 'to remove' THEN 'to upgrade'
                        WHEN state = 'uninstalled' THEN 'to install'
                        ELSE state
                    END,
                    demo=
          (SELECT demo
           FROM ir_module_module
           WHERE name='base')
        FROM deps d
        WHERE m.id = d.mod_id {subquery} RETURNING m.name,
                                         m.state;

    """
    ).format(**_params)
    cr.execute(_query, locals())
    state = dict(cr.fetchall()).get(module)
    return state


def new_module_dep(cr, module, new_dep):
    """Add a new dependency to a module."""
    # One new dep at a time
    _query = sql.SQL(
        """

        INSERT INTO ir_module_module_dependency(name, module_id)
        SELECT %(new_dep)s,
               id
        FROM ir_module_module m
        WHERE name = %(module)s
          AND NOT EXISTS
            (SELECT 1
             FROM ir_module_module_dependency
             WHERE module_id = m.id
               AND name = %(new_dep)s);

    """
    )
    cr.execute(_query, locals())
    _query = sql.SQL(
        """

        SELECT state
        FROM ir_module_module
        WHERE name = %(module)s;

    """
    )
    cr.execute(_query, locals())

    mod_state = (cr.fetchone() or ["n/a"])[0]
    if mod_state in _INSTALLED_MODULE_STATES:
        # Module was installed, need to install all its deps, recursively,
        # to make sure the new dep is installed
        force_install_module(cr, module)


def remove_module_deps(cr, module, old_deps):
    """Remove a dependencies from a module.
        :param str module: name of the module whose dependencies are removed
        :param tuple old_deps: list of dependencies to be removed
    """
    assert isinstance(old_deps, tuple)
    _query = sql.SQL(
        """

        DELETE
        FROM ir_module_module_dependency
        WHERE module_id =
            (SELECT id
             FROM ir_module_module
             WHERE name = %(module)s)
          AND name IN %(old_deps)s

    """
    )
    cr.execute(_query, locals())


def module_deps_diff(cr, module, add=(), remove=()):
    """Change the dependencies of a module (adding and removing).

        :param str module: technical name of the module with dependency changes
        :param list add: modules to add as dependencies
        :param list remove: modules to remove as dependencies
    """
    for new_dep in add:
        new_module_dep(cr, module, new_dep)
    if remove:
        remove_module_deps(cr, module, tuple(remove))


def new_module(cr, module, deps=(), auto_install=False):
    """Make a new module known to the database.

        :param str module: technical name of the new module
        :param list deps: modules to add as dependencies
        :param bool auto_install: whether the module will auto install if all
                                  its dependencies are installed
    """
    _query = sql.SQL(
        """

        SELECT count(1)
        FROM ir_module_module
        WHERE name = %(module)s

    """
    )
    cr.execute(_query, locals())
    if cr.fetchone()[0]:
        # Avoid duplicate entries for module which is already installed,
        # even before it has become standard module in new version
        # Also happen for modules added afterward, which should be added by multiple series.
        return

    if deps and auto_install:
        state = "to install" if modules_installed(cr, *deps) else "uninstalled"
    else:
        state = "uninstalled"
    _query = sql.SQL(
        """

        INSERT INTO ir_module_module (name, STATE, demo)
        VALUES (%(module)s, %(state)s,
                  (SELECT demo
                   FROM ir_module_module
                   WHERE name='base')) RETURNING id

    """
    )
    cr.execute(_query, locals())
    new_id, = cr.fetchone()

    _query = sql.SQL(
        """

        INSERT INTO ir_model_data (name, module, noupdate, model, res_id)
        VALUES ('module_'||%(module)s,
                                    'base',
                                    't',
                                    'ir.module.module',
                                    %(new_id)s)

    """
    )
    cr.execute(_query, locals())

    for dep in deps:
        new_module_dep(cr, module, dep)


def force_migration_of_fresh_module(cr, module):
    """Force migration scripts to be executed for new modules.

    In some cases, new (or forced installed) modules need a migration script to grab data
    form other modules. In that case, forcing its state to be 'to upgrade' will allow
    migration scripts to be run even if the module was not installed before the migration.
    """
    filename, _ = frame_codeinfo(currentframe(), 1)
    version = ".".join(filename.split(os.path.sep)[-2].split(".")[:2])

    # Force module state to be in `to upgrade`.
    # Needed for migration script execution. See http://git.io/vnF7f
    _query = sql.SQL(
        """

        UPDATE ir_module_module
        SET state = 'to upgrade',
                    latest_version = %(version)s
        WHERE name = %(module)s
          AND state = 'to install' RETURNING id

    """
    )
    cr.execute(_query, locals())
    if cr.rowcount:
        # Force module in `init` mode beside its state is forced to `to upgrade`
        # See http://git.io/vnF7O
        odoo.tools.config["init"][module] = "oh yeah!"


def split_group(cr, from_groups, to_group):
    """Make users members of all `from_groups` members of `to_group`"""

    def check_group(g):
        if isinstance(g, str):
            gid = ref(cr, g)
            if not gid:
                _logger.warning("split_group(): Unknow group: %r", g)
            return gid
        return g

    if not isinstance(from_groups, (list, tuple, set)):
        from_groups = [from_groups]

    from_groups = [g for g in map(check_group, from_groups) if g]
    if not from_groups:
        return

    if isinstance(to_group, str):
        to_group = ref(cr, to_group)

    assert to_group

    _query = sql.SQL(
        """

        INSERT INTO res_groups_users_rel(uid, gid)
        SELECT uid,
               %(to_group)s
        FROM res_groups_users_rel
        GROUP BY uid HAVING array_agg(gid) @> %(from_groups)s EXCEPT
        SELECT uid,
               gid
        FROM res_groups_users_rel
        WHERE gid = %(to_group)s

    """
    )
    cr.execute(_query, locals())


# Models & Fields utilities


def create_m2m(cr, m2m, fk1, fk2, col1=None, col2=None):
    """Create a many2many relation table.

        :param str m2m: relation table name
        :param str fk1: table referenced by first column
        :param str fk2: table referenced by second column
        :param str col1: first column name
        :param str col2: second column name
    """
    if col1 is None:
        col1 = "%s_id" % fk1
    if col2 is None:
        col2 = "%s_id" % fk2

    _params = {
        "m2m": sql.Identifier(m2m),
        "col1": sql.Identifier(col1),
        "fk1": sql.Identifier(fk1),
        "col2": sql.Identifier(col2),
        "fk2": sql.Identifier(fk2),
    }
    _query = sql.SQL(
        """

        CREATE TABLE {m2m}({col1} integer NOT NULL REFERENCES {fk1}(id) ON
                           DELETE CASCADE, {col2} integer NOT NULL REFERENCES {fk2}(id) ON
                           DELETE CASCADE,
                                  UNIQUE ({col1}, {col2}));


        CREATE INDEX ON {m2m}({col1});


        CREATE INDEX ON {m2m}({col2});
    """
    ).format(**_params)
    cr.execute(_query)


def ensure_m2o_func_field_data(cr, src_table, column, dst_table):
    """Fix broken many2one relations.

        If any `column` not present in `dst_table`, remove column from `src_table` in
        order to force recomputation of the function field

        WARN: only call this method on m2o function/related fields!!
    """
    if not column_exists(cr, src_table, column):
        return
    _params = {
        "src_table": sql.Identifier(src_table),
        "column": sql.Identifier(column),
        "dst_table": sql.Identifier(dst_table),
    }
    _query = sql.SQL(
        """
        SELECT count(1)
        FROM {src_table}
        WHERE {column} NOT IN
            (SELECT id
             FROM {dst_table})
    """
    ).format(**_params)
    cr.execute(_query)
    if cr.fetchone()[0]:
        remove_column(cr, src_table, column, cascade=True)


def uniq_tags(cr, model, uniq_column="name", order="id"):
    """Deduplicate "tag" models entries.

        Should only be referenced as many2many

        By using `uniq_column=lower(name)` and `order=name`
        you can prioritize tags in CamelCase/UPPERCASE.
    """
    table = table_of_model(cr, model)
    _upds = sql.SQL("")
    for ft, fc, _, da in get_fk(cr, table):
        assert da == "c"  # should be a ondelete=cascade fk
        cols = get_columns(cr, ft, ignore=(fc,))[0]
        assert len(cols) == 1  # it's a m2, should have only 2 columns

        _params = {
            "rel": sql.Indentifier(ft),
            "c1": sql.Identifier(cols[0]),
            "c2": sql.Identifier(cols[1]),
        }

        _upds += sql.SQL(
            """

            INSERT INTO {rel}({c1}, {c2})
            SELECT r.{c1}, d.id
            FROM {rel} r
            JOIN dups d ON (r.{c2} = ANY(d.others)) EXCEPT
            SELECT r.{c1}, r.{c2}
            FROM {rel} r
            JOIN dups d ON (r.{c2} = d.id)

        """
        ).format(**_params)

    assert _upds  # if not m2m found, there is something wrong...

    # TODO: There is an issue here, find out teleologically, what was meant.
    updates = sql.SQL(", ".join("_upd_%s AS (%s)" % x for x in enumerate(_upds)))
    _params = {
        "table": sql.Identifier(table),
        "uniq_column": sql.Identifier(uniq_column),
        "updates": updates,
    }
    _query = sql.SQL(
        """

         WITH dups AS
          ( SELECT (array_agg(id ORDER BY {order}))[1] AS id,
                   (array_agg(id ORDER BY {order}))[2:array_length(array_agg(id), 1)] AS others
           FROM {table}
           GROUP BY {uniq_column} HAVING count(id) > 1 ), _upd_imd AS
          ( UPDATE ir_model_data x
           SET res_id = d.id
           FROM dups d
           WHERE x.model = %(model)s
             AND x.res_id = ANY(d.others) ), {updates}
        DELETE
        FROM {table}
        WHERE id IN (
          SELECT unnest(others)
          FROM dups)

    """
    ).format(**_params)
    cr.execute(_query, locals())


def remove_field(cr, model, fieldname, cascade=False):
    """Remove a field.

        :param str model: name of the field's model
        :param str fieldname: name of the field (...)
        :param bool cascade: if True, all records having a FKEY pointing to this field
                             will be cascade-deleted (default: False)
    """
    if fieldname == "id":
        # called by `remove_module`. May happen when a model defined in a removed module was
        # overwritten by another module in previous versions.
        return remove_model(cr, model)

    # clean dashboards' `group_by`
    _query = sql.SQL(
        """

        SELECT array_agg(f.name),
               array_agg(aw.id)
        FROM ir_model_fields f
        JOIN ir_act_window aw ON aw.res_model = f.model
        WHERE f.model = %s
          AND f.name = %s
        GROUP BY f.model

    """
    )
    cr.execute(_query, locals())
    for fields, actions in cr.fetchall():
        _query = sql.SQL(
            """

            SELECT id,
                   arch
            FROM ir_ui_view_custom
            WHERE arch ~ %s

        """
        )
        cr.execute(_query, ["name=[\"'](%s)[\"']" % "|".join(map(str, actions))])
        for id, arch in ((x, lxml.etree.fromstring(y)) for x, y in cr.fetchall()):
            for action in arch.iterfind(".//action"):
                context = eval(action.get("context", "{}"), UnquoteEvalContext())
                if context.get("group_by"):
                    context["group_by"] = list(set(context["group_by"]) - set(fields))
                    action.set("context", str(context))
            _query = sql.SQL(
                """

                UPDATE ir_ui_view_custom
                SET arch = %s
                WHERE id = %s

            """
            )
            cr.execute(_query, [lxml.etree.tostring(arch, encoding="unicode"), id])

    _query = sql.SQL(
        """

        DELETE
        FROM ir_model_fields
        WHERE model=%(model)s
          AND name=%(fieldname)s RETURNING id

    """
    )
    cr.execute(_query, locals())
    fids = tuple(vals[0] for vals in cr.fetchall())
    if fids:
        _query = sql.SQL(
            """

            DELETE
            FROM ir_model_data
            WHERE model = 'ir.model.fields'
              AND res_id IN %(fids)s

        """
        )
        cr.execute(_query, **locals())

    # cleanup translations
    _query = sql.SQL(
        """

        DELETE
        FROM ir_translation
        WHERE name = %s
          AND TYPE IN ('field',
                       'help',
                       'model',
                       'selection') -- ignore wizard_* translations

    """
    )
    cr.execute(_query, ["{},{}".format(model, fieldname)])

    table = table_of_model(cr, model)
    remove_column(cr, table, fieldname, cascade=cascade)


def move_field_to_module(cr, model, fieldname, old_module, new_module):
    """Move a field to another module."""
    name = IMD_FIELD_PATTERN % (model.replace(".", "_"), fieldname)

    _query = sql.SQL(
        """

        UPDATE ir_model_data
        SET module=%(new_module)s
        WHERE model='ir.model.fields'
          AND name=%(name)s
          AND module=%(old_module)s

    """
    )
    cr.execute(_query, locals())


def rename_field(cr, model, old, new, update_references=True):
    """Rename a module. Yes, really.

        :param bool update_references: if True, references to that field in
                                       filters, saved exports, etc. will be
                                       adapted (default: True)
    """

    _query = sql.SQL(
        """

        UPDATE ir_model_fields
        SET name = %(new)s
        WHERE model = %(model)s
          AND name = %(old)s RETURNING id

    """
    )
    cr.execute(_query, locals())
    [fid] = cr.fetchone() or [None]
    if fid:
        name = IMD_FIELD_PATTERN % (model.replace(".", "_"), new)

        _query = sql.SQL(
            """

            UPDATE ir_model_data
            SET name=%(name)s
            WHERE model='ir.model.fields'
              AND res_id=%(fid)s;


            UPDATE ir_property
            SET name=%(new)s
            WHERE fields_id=%(fid)s;

        """
        )
        cr.execute(_query, locals())
    _query = sql.SQL(
        """


        UPDATE ir_translation
        SET name = %s
        WHERE name = %s
          AND TYPE IN ('field',
                       'help',
                       'model',
                       'selection') -- ignore wizard_* translations

    """
    )
    cr.execute(_query, ["{},{}".format(model, new), "{},{}".format(model, old)])
    table = table_of_model(cr, model)
    # NOTE table_exists is needed to avoid altering views
    if table_exists(cr, table) and column_exists(cr, table, old):
        _params = {
            "table": sql.Identifier(table),
            "old": sql.Identifier(old),
            "new": sql.Identifier(new),
        }
        _query = sql.SQL(
            """

            ALTER TABLE {table} RENAME COLUMN {old} TO {new}

        """
        ).format(**_params)
        cr.execute(_query)

    if update_references:
        update_field_references(cr, old, new, only_models=(model,))


def make_field_company_dependent(
    cr,
    model,
    field,
    field_type,
    target_model=None,
    default_value=None,
    default_value_ref=None,
    company_field="company_id",
):
    """Convert a field to be company dependent (old `property` field attributes).

    Notes:
        `target_model` is only use when `type` is "many2one".
        The `company_field` can be an sql expression.
        You may use `t` to refer the model's table.
    """
    type2field = {
        "char": "value_text",
        "float": "value_float",
        "boolean": "value_integer",
        "integer": "value_integer",
        "text": "value_text",
        "binary": "value_binary",
        "many2one": "value_reference",
        "date": "value_datetime",
        "datetime": "value_datetime",
        "selection": "value_text",
    }

    assert field_type in type2field
    value_field = type2field[field_type]

    _query = sql.SQL(
        """

        SELECT id
        FROM ir_model_fields
        WHERE model = %(model)s
          AND name = %(field)s

    """
    )
    cr.execute(_query, locals())
    [fields_id] = cr.fetchone()

    table = table_of_model(cr, model)

    _params = {
        "field": sql.Identifier(field),
        "value_field": sql.Identifier(value_field),
        "company_field": sql.Identifier(company_field),
        "table": sql.Identifier(table),
        "target_model_prefix": sql.Literal("{},".format(target_model)),
        "model_prefix": sql.Literal("{},".format(model)),
    }

    if default_value is None:
        where_clause = sql.SQL(
            """

            {field} IS NOT NULL

        """
        ).format(**_params)
    else:
        where_clause = sql.SQL(
            """

            {field} != %(default_value)s

        """
        ).format(**_params)

    if field_type != "many2one":
        value_select = sql.Identifier(field)
    else:
        # for m2o, the store value is a reference field, so in format `model,id`
        value_select = sql.SQL(
            """

        CONCAT({target_model_prefix}, {field})

        """
        ).format(**_params)

    _params = dict(
        {"value_select": value_select, "where_clause": where_clause}, **_params
    )
    # TODO: remove me when anonimization module is removed
    if is_field_anonymized(cr, model, field):
        # if field is anonymized, we need to create a property for each record
        where_clause = "true"
        # and we need to unanonymize its values
        ano_default_value = cr.mogrify("%s", [default_value])
        if field_type != "many2one":
            ano_value_select = "%(value)s"
        else:
            ano_value_select = sql.SQL(
                """

            CONCAT({target_model_prefix}, %(value)s)

            """
            ).format(**_params)

        register_unanonymization_query(
            cr,
            model,
            field,
            """
            UPDATE ir_property
               SET {value_field} = CASE WHEN %(value)s IS NULL THEN {ano_default_value}
                                        ELSE {ano_value_select} END
             WHERE res_id = CONCAT({model_prefix}, %(id)s)
               AND name='{field}'
               AND type='{field_type}'
               AND fields_id={fields_id}
            """.format(
                **locals()
            ),
        )

    _query = sql.SQL(
        """

         WITH cte AS
          ( SELECT CONCAT({model_prefix}, id) AS res_id,
                   {value_select} AS value,
                   ({company_field})::integer AS company
           FROM {table} t
           WHERE {where_clause} )
        INSERT INTO ir_property(name, type, fields_id, company_id, res_id, {value_field})
        SELECT %(field)s,
               %(field_type)s,
               %(fields_id)s,
               cte.company,
               cte.res_id,
               cte.value
        FROM cte
        WHERE NOT EXISTS
            (SELECT 1
             FROM ir_property
             WHERE fields_id = %(fields_id)s
               AND COALESCE(company_id, 0) = COALESCE(cte.company, 0)
               AND res_id=cte.res_id)

    """
    ).format(**_params)
    cr.execute(_query, locals())
    # default property
    if default_value:
        _query = sql.SQL(
            """

            INSERT INTO ir_property(name, type, fields_id, {value_field})
            VALUES (%(field)s,
                    %(field_type)s,
                    %(fields_id)s,
                    %(default_value)s) RETURNING id

        """
        )
        cr.execute(_query, locals())
        [prop_id] = cr.fetchone()
        if default_value_ref:
            module, _, xid = default_value_ref.partition(".")

            _query = sql.SQL(
                """

                INSERT INTO ir_model_data (module, name, model, res_id, noupdate)
                VALUES (%(module)s,
                        %(xid)s,
                        'ir.property',
                        %(prop_id)s,
                        TRUE)

            """
            )
            cr.execute(_query, locals())

    remove_column(cr, table, field, cascade=True)


def is_field_anonymized(cr, model, field):
    """Check if a field has been anonymized prior to the migration."""
    if not module_installed(cr, "anonymization"):
        return False
    _query = sql.SQL(
        """

        SELECT id
        FROM ir_model_fields_anonymization
        WHERE model_name = %(model)s
          AND field_name = %(field)s
          AND STATE = 'anonymized'

    """
    )
    cr.execute(_query, locals())
    return bool(cr.rowcount)


def update_field_references(cr, old, new, only_models=None):
    """Replaces references to a field in several crash-prone places.

    Replace all references to field `old` to `new` in:
        - ir_filters
        - ir_exports_line
        - ir_act_server
        - ir_rule
        - mail.mass_mailing

        :param list only_models: list of models affected by the fieldname change
    """
    p = {
        "old": r"\y{}\y".format(old),
        "new": new,
        "def_old": r"\ydefault_{}\y".format(old),
        "def_new": "default_{}".format(new),
        "models": tuple(only_models) if only_models else (),
    }

    _query = sql.SQL(
        """

        UPDATE ir_filters
        SET domain = regexp_replace(domain, %(old)s, %(new)s, 'g'),
            context = regexp_replace(regexp_replace(context, %(old)s, %(new)s, 'g'), %(def_old)s, %(def_new)s, 'g')

    """
    )
    if column_exists(cr, "ir_filters", "sort"):
        _query += sql.SQL(", sort = regexp_replace(sort, %(old)s, %(new)s, 'g')")

    if only_models:
        _query += sql.SQL(" WHERE model_id IN %(models)s")

    cr.execute(_query, p)

    # ir.exports.line
    _query = sql.SQL(
        """

        UPDATE ir_exports_line l
        SET name = regexp_replace(l.name, %(old)s, %(new)s, 'g')

    """
    )
    if only_models:
        _query += sql.SQL(
            """

            FROM ir_exports e
            WHERE e.id = l.export_id
              AND e.resource IN %(models)s

        """
        )
    cr.execute(_query, p)

    # ir.action.server
    col_prefix = ""
    if not column_exists(cr, "ir_act_server", "condition"):
        col_prefix = "--"  # sql comment the line
    _query = sql.SQL(
        """

        UPDATE ir_act_server s
        SET {col_prefix} condition = regexp_replace(condition, %(old)s, %(new)s, 'g'),
            code = regexp_replace(code, %(old)s, %(new)s, 'g')

    """.format(
            col_prefix=col_prefix
        )
    )
    if only_models:
        _query += sql.SQL(
            """

            FROM ir_model m
            WHERE m.id = s.model_id
              AND m.model IN %(models)s
              AND

        """
        )
    else:
        _query += sql.SQL(" WHERE ")

    _query += sql.SQL("s.state = 'code'")
    cr.execute(_query, p)

    # ir.rule
    _query = sql.SQL(
        """

        UPDATE ir_rule r
        SET domain_force = regexp_replace(domain_force, %(old)s, %(new)s, 'g')

    """
    )
    if only_models:
        _query += sql.SQL(
            """

            FROM ir_model m
            WHERE m.id = r.model_id
              AND m.model IN %(models)s

        """
        )
    cr.execute(_query, p)

    # mass mailing
    if column_exists(cr, "mail_mass_mailing", "mailing_domain"):
        _query = sql.SQL(
            """

            UPDATE mail_mass_mailing u
            SET mailing_domain = regexp_replace(u.mailing_domain, %(old)s, %(new)s, 'g')

        """
        )
        if only_models:
            if column_exists(cr, "mail_mass_mailing", "mailing_model_id"):
                _query += sql.SQL(
                    """

                    FROM ir_model m
                    WHERE m.id = u.mailing_model_id
                      AND m.model IN %(models)s

                """
                )
            else:
                _query += sql.SQL("WHERE u.mailing_model IN %(models)s")
        cr.execute(_query, p)


def recompute_fields(cr, model, fields, ids=None, logger=_logger, chunk_size=256):
    """Recompute fields using the ORM.

        :param str model:
        :param list fields:
        :param list ids: list of ids of records on which the recompute will be done;
                         if not set, will recompute for all records
        :param logger: logger used during the processing (default: current logger)
        :param integer chunk_size: batch size for recomputing (default: 256)
    """
    if ids is None:

        _query = sql.SQL(
            """

            SELECT id FROM "%(table)s"

        """
        )
        cr.execute(_query, dict(table=table_of_model(cr, model)))
        ids = tuple(map(itemgetter(0), cr.fetchall()))

    Model = env(cr)[model]
    size = (len(ids) + chunk_size - 1) / chunk_size
    qual = "%s %d-bucket" % (model, chunk_size) if chunk_size != 1 else model
    for subids in log_progress(
        chunks(ids, chunk_size, list), qualifier=qual, logger=logger, size=size
    ):
        records = Model.browse(subids)
        for field in fields:
            records._recompute_todo(records._fields[field])
        records.recompute()
        records.invalidate_cache()


def fix_wrong_m2o(cr, table, column, target, value=None):
    """Fix missing foreign keys references.

        :param str table: table to correct
        :param str column: column to correct
        :param str target: destination table of the FKEY
        :param value: value to set instead of the missing foreign key reference
                      that will be parsed by psycopg2 (variable type)
                      (default: None - NULL)
    """
    _params = {
        "table": sql.Identifier(table),
        "target": sql.Identifier(target),
        "column": sql.Identifier(column),
    }
    _query = sql.SQL(
        """

         WITH wrongs_m2o AS
          ( SELECT s.id
           FROM {table} s
           LEFT JOIN {target} t ON s.{column} = t.id
           WHERE s.{column} IS NOT NULL
             AND t.id IS NULL )
        UPDATE {table} s
        SET {column} = %(value)s
        FROM wrongs_m2o w
        WHERE s.id = w.id

    """
    ).format(**_params)
    cr.execute(_query, locals())


def remove_model(cr, model, drop_table=True):
    """Remove a model 😉."""
    model_underscore = model.replace(".", "_")

    # remove references
    for ir in indirect_references(cr):
        if ir.table == "ir_model":
            continue
        _params = {
            "table": sql.Identifier(ir.table),
            "where_clause": sql.SQL(ir.model_filter(placeholder="%(model)s")),
        }
        _query = sql.SQL(
            """

            DELETE
            FROM {table}
            WHERE {where_clause} RETURNING id

        """
        ).format(**_params)
        cr.execute(_query, locals())
        ids = tuple(vals[0] for vals in cr.fetchall())
        remove_refs(cr, model_of_table(cr, ir.table), ids)

    remove_refs(cr, model)

    _query = sql.SQL(
        """

        SELECT id
        FROM ir_model
        WHERE model = %(model)s

    """
    )
    cr.execute(_query, locals())
    [mod_id] = cr.fetchone() or [None]
    if mod_id:
        # some required fk are "ON DELETE SET NULL".
        for tbl in "base_action_rule google_drive_config".split():
            if column_exists(cr, tbl, "model_id"):
                _params = {"tbl": sql.Identifier(tbl)}
                _query = sql.SQL(
                    """

                    DELETE
                    FROM {tbl}
                    WHERE model_id = %(mod_id)s

                """
                ).format(**_params)
                cr.execute(_query, locals())

        _query = sql.SQL(
            """


            DELETE
            FROM ir_model_constraint
            WHERE model=%(mod_id)s;


            DELETE
            FROM ir_model_relation
            WHERE model=%(mod_id)s;

             --- Drop XML IDs of ir.rule and ir.model.access records that will be cascade-dropped,
             --- when the ir.model record is dropped - just in case they need to be re-created

            DELETE
            FROM ir_model_data x USING ir_rule a
            WHERE x.res_id = a.id
              AND x.model='ir.rule'
              AND a.model_id = %(mod_id)s;


            DELETE
            FROM ir_model_data x USING ir_model_access a
            WHERE x.res_id = a.id
              AND x.model='ir.model.access'
              AND a.model_id = %(mod_id)s;


            DELETE
            FROM ir_model
            WHERE id=%(mod_id)s;

        """
        )
        cr.execute(_query, locals())

    _query = sql.SQL(
        """

        DELETE
        FROM ir_model_data
        WHERE model='ir.model'
          AND name = %(name)s;

        DELETE
        FROM ir_model_data
        WHERE model='ir.model.fields'
          AND name LIKE %(name_like)s;

    """
    )
    cr.execute(
        _query,
        dict(
            name="model_{}".format(model_underscore),
            name_like=(IMD_FIELD_PATTERN % (model_underscore, "%")).replace("_", r"\_"),
        ),
    )

    table = table_of_model(cr, model)
    _params = {"table": sql.Identifier(table)}
    if drop_table:
        if table_exists(cr, table):
            _query = sql.SQL(
                """

                DROP TABLE {table} CASCADE

            """
            ).format(**_params)
            cr.execute(_query)
        elif view_exists(cr, table):
            # For auto=False models...
            _query = sql.SQL(
                """

                DROP VIEW {table} CASCADE

            """
            ).format(**_params)
            cr.execute(_query)


def remove_refs(cr, model, ids=None):
    """Remove non-sql enforced references pointing to the specified model.

    e.g. reference fields, translations, ...
    """
    if ids is None:
        match = sql.SQL("like %(needle)s")
        needle = model + ",%"
    else:
        if not ids:
            return
        match = sql.SQL("in %(needle)s")
        needle = tuple("{},{}".format(model, i) for i in ids)

    # "model-comma" fields
    _query = sql.SQL(
        """

        SELECT model,
               name
        FROM ir_model_fields
        WHERE ttype = 'reference'
        UNION
        SELECT 'ir.translation',
               'name'

    """
    )
    cr.execute(_query)

    for ref_model, ref_column in cr.fetchall():
        table = table_of_model(cr, ref_model)
        # NOTE table_exists is needed to avoid deleting from views
        if table_exists(cr, table) and column_exists(cr, table, ref_column):
            _params = {
                "table": sql.Identifier(table),
                "ref_column": sql.Identifier(ref_column),
                "match": match,
            }
            query_tail = sql.SQL(
                """

                FROM {table}
                WHERE {ref_column} {match}

            """
            ).format(**_params)
            if ref_model == "ir.ui.view":
                _query = sql.SQL("""SELECT id """) + query_tail
                cr.execute(_query, locals())
                for (view_id,) in cr.fetchall():
                    remove_view(
                        cr, view_id=view_id, deactivate_custom=True, silent=True
                    )
            elif ref_model == "ir.ui.menu":
                _query = sql.SQL("""SELECT id """) + query_tail
                cr.execute(_query, locals())
                menu_ids = tuple(m[0] for m in cr.fetchall())
                remove_menus(cr, menu_ids)
            else:
                _query = sql.SQL("""DELETE """) + query_tail
                cr.execute(_query, locals())
                # TODO make it recursive?

    if table_exists(cr, "ir_values"):
        column, _ = _ir_values_value(cr)
        _params = {"column": sql.SQL(column), "match": match}

        _query = sql.SQL(
            """

            DELETE
            FROM ir_values
            WHERE {column} {match}

        """
        ).format(**_params)
        cr.execute(_query, locals())

    if ids is None:
        _query = sql.SQL(
            """

            DELETE
            FROM ir_translation
            WHERE name = %(model)s
              AND TYPE IN ('constraint',
                           'sql_constraint',
                           'view',
                           'report',
                           'rml',
                           'xsl')

        """
        )
        cr.execute(_query, locals())


def move_model(cr, model, from_module, to_module, move_data=False, delete=False):
    """Move model `model` from `from_module` to `to_module`.

        :param bool move_data: reassign all xmlids from the `from_module`
                               referencing `model` records to `to_module`
                               (default: False)
        :param bool delete: if True and `to_module` is not installed,
                            delete the model (default: False)
    """
    if delete and not module_installed(cr, to_module):
        remove_model(cr, model)
        return

    model_underscore = model.replace(".", "_")
    name = "model_%s" % model_underscore
    name_like = (IMD_FIELD_PATTERN % (model_underscore, "%")).replace("_", r"\_")
    _query = sql.SQL(
        """

        UPDATE ir_model_data
        SET module = %(to_module)s
        WHERE module = %(from_module)s
          AND model = 'ir.model'
          AND name = %(name)s;


        UPDATE ir_model_data
        SET module = %(to_module)s
        WHERE module = %(from_module)s
          AND model='ir.model.fields'
          AND name LIKE %(name)s;

    """
    )
    cr.execute(_query, locals())

    if move_data:
        _query = sql.SQL(
            """

            UPDATE ir_model_data
            SET module = %(to_module)s
            WHERE module = %(from_module)s
              AND model = %(model)s

        """
        )
        cr.execute(_query, locals())


def rename_model(cr, old, new, rename_table=True):
    """Rename a model.

        :param bool rename_table: if True, the table will be renamed according
                                  to the Odoo ORM (e.g. `ir.rule` -> `ir_rule`)
    """
    if rename_table:
        old_table = table_of_model(cr, old)
        new_table = table_of_model(cr, new)
        _params = {
            "old_table": sql.Identifier(old_table),
            "new_table": sql.Identifier(new_table),
            "old_table_seq_id": sql.Identifier("{}_id_seq".format(old_table)),
            "new_table_sqe_id": sql.Identifier("{}_id_seq".format(new_table)),
        }
        _query = sql.SQL(
            """

            ALTER TABLE {old_table} RENAME TO {new_table};


            ALTER SEQUENCE {old_table_seq_id} RENAME TO {new_table_sqe_id};

        """
        ).format(**_params)
        cr.execute(_query, locals())

        # find & rename primary key, may still use an old name from a former migration
        _query = sql.SQL(
            """

            SELECT conname
            FROM pg_index,
                 pg_constraint
            WHERE indrelid = %(new_table)s::regclass
              AND indisprimary
              AND conrelid = indrelid
              AND conindid = indexrelid
              AND confrelid = 0;

        """
        )
        cr.execute(_query, locals())

        primary_key, = cr.fetchone()

        _params = dict(
            {
                "primary_key": sql.Identifier(primary_key),
                "new_table_pkey": sql.Identifier("{}_pkey".format(new_table)),
            },
            **_params
        )
        _query = sql.SQL(
            """

            ALTER INDEX {primary_key} RENAME TO {new_table_pkey}

        """
        ).format(**_params)
        cr.execute(_query, locals())

        # DELETE all constraints and indexes (ignore the PK), ORM will recreate them.
        _query = sql.SQL(
            """

            SELECT CONSTRAINT_NAME
            FROM information_schema.table_constraints
            WHERE TABLE_NAME = %(new_table)s
              AND constraint_type != 'PRIMARY KEY'
              AND constraint_name !~ '^[0-9_]+_not_null$'

        """
        )
        cr.execute(_query, locals())
        for (constrain,) in cr.fetchall():
            _params = {
                "constrain": sql.Identifier(constrain),
                "new_table": sql.Identifier(new_table),
            }
            _query = sql.SQL(
                """

                DELETE
                FROM ir_model_constraint
                WHERE name = %(constrain)s;


                ALTER TABLE {new_table}
                DROP CONSTRAINT {constrain};

            """
            ).format(**_params)
            cr.execute(_query, locals())

    updates = [r[:2] for r in res_model_res_id(cr)]

    for model, column in updates:
        table = table_of_model(cr, model)
        _params = {"table": sql.Identifier(table), "column": sql.Identifier(column)}
        _query = sql.SQL(
            """

            UPDATE {table}
            SET {column} = %(new)s
            WHERE {column} = %(old)s

        """
        ).format(**_params)
        cr.execute(_query, locals())

    # "model-comma" fields
    _query = sql.SQL(
        """

        SELECT model,
               name
        FROM ir_model_fields
        WHERE ttype = 'reference'
        UNION
        SELECT 'ir.translation',
               'name'

    """
    )
    cr.execute(_query, locals())
    old_like = "{},%".format(old)
    substr_from = len(old)
    for model, column in cr.fetchall():
        table = table_of_model(cr, model)
        if column_exists(cr, table, column):
            _params = {"table": sql.Identifier(table), "column": sql.Identifier(column)}

            _query = sql.SQL(
                """

                UPDATE {table}
                SET {column} = %(new)s || substring({column} FROM %(substr_from)s)
                WHERE {column} LIKE %(old_like)s

            """
            ).format(**_params)
            cr.execute(_query, locals())

    if table_exists(cr, "ir_values"):
        column_read, cast_write = _ir_values_value(cr)
        _params = {
            "cast0": sql.SQL(cast_write.partition("%s")[0]),
            "cast2": sql.SQL(cast_write.partition("%s")[2]),
            "column": sql.SQL(column_read),
        }

        _query = sql.SQL(
            """

            UPDATE ir_values
            SET value = {cast0}%(new)s || substring({column} FROM %(substr_from)s){cast2}
            WHERE {column} LIKE %(old_like)s

        """
        ).format(**_params)
        cr.execute(_query, locals())

    old_underscore = old.replace(".", "_")
    new_underscore = new.replace(".", "_")
    new_name = "model_%s" % new_underscore
    old_name = "model_%s" % old_underscore
    new_field_name = "field_%s" % new_underscore
    old_field_prefix_length = len(old_underscore) + 6
    name_like = (IMD_FIELD_PATTERN % (old_underscore, "%")).replace("_", r"\_")
    _query = sql.SQL(
        """

        UPDATE ir_translation
        SET name = %(new)s
        WHERE name = %(old)s
          AND TYPE IN ('constraint',
                       'sql_constraint',
                       'view',
                       'report',
                       'rml',
                       'xsl');


        UPDATE ir_model_data
        SET name = %(new_name)s
        WHERE model = 'ir.model'
          AND name = %(old_name)s;


        UPDATE ir_model_data
        SET name = %(new_field_name)s || substring(name FROM %(old_field_prefix_length)s)
        WHERE model = 'ir.model.fields'
          AND name LIKE %(name_like)s;

    """
    )
    cr.execute(_query, locals())

    col_prefix = ""
    if not column_exists(cr, "ir_act_server", "condition"):
        col_prefix = "--"  # sql comment the line

    _params = {
        "col_prefix": sql.SQL(col_prefix),
        "old": sql.SQL(old),
        "new": sql.SQL(new),
    }
    _query = sql.SQL(
        r"""

        UPDATE ir_act_server
        -- regex matching model name wrapped by quotes eg in env['model.name'] or env["model.name"]
        SET {col_prefix} condition = regexp_replace(condition, '([''"]){old}\1', '\1{new}\1', 'g'),
                                   code = regexp_replace(code, '([''"]){old}\1', '\1{new}\1', 'g')
    """
    ).format(**_params)
    cr.execute(_query, locals())


def replace_record_references(cr, old, new, replace_xmlid=True):
    """Replace all (in)direct references of a record by another"""
    assert isinstance(old, tuple) and len(old) == 2
    assert isinstance(new, tuple) and len(new) == 2

    if not old[1]:
        return

    return replace_record_references_batch(
        cr, {old[1]: new[1]}, old[0], new[0], replace_xmlid
    )


def replace_record_references_batch(
    cr, id_mapping, model_src, model_dst=None, replace_xmlid=True
):
    assert id_mapping
    assert all(isinstance(v, int) and isinstance(k, int) for k, v in id_mapping.items())

    if model_dst is None:
        model_dst = model_src

    old = tuple(id_mapping.keys())
    new = tuple(id_mapping.values())
    jmap = json.dumps(id_mapping)

    def genmap(fmt_k, fmt_v=None):
        # generate map using given format
        fmt_v = fmt_k if fmt_v is None else fmt_v
        m = {fmt_k % k: fmt_v % v for k, v in id_mapping.items()}
        return json.dumps(m), tuple(m.keys())

    if model_src == model_dst:
        # 7 time faster than using pickle.dumps
        pmap, pmap_keys = genmap("I%d\n.")
        smap, smap_keys = genmap("%d")

        column_read, cast_write = _ir_values_value(cr)

        for table, fk, _, _ in get_fk(cr, table_of_model(cr, model_src)):
            _params = {"table": sql.Identifier(table), "fk": sql.Identifier(fk)}

            jmap_query = sql.SQL(
                """

                UPDATE {table} t
                SET {fk} = (%(jmap)s::json->>{fk}::varchar)::int4
                WHERE {fk} IN %(old)s

            """
            ).format(**_params)

            col2 = None
            if not column_exists(cr, table, "id"):
                # seems to be a m2m table. Avoid duplicated entries
                cols = get_columns(cr, table, ignore=(fk,))[0]
                assert len(cols) == 1  # it's a m2, should have only 2 columns
                col2 = cols[0]
                _params = {
                    "table": sql.Identifier(table),
                    "fk": sql.Identifier(fk),
                    "col2": sql.Identifier(col2),
                    "jmap_query": jmap_query,
                }

                _query = sql.SQL(
                    """

                     WITH _existing AS
                      ( SELECT {col2}
                       FROM {table}
                       WHERE {fk} IN %(new)s ) {jmap_query}
                    AND NOT EXISTS
                      (SELECT 1
                       FROM _existing
                       WHERE {col2}=t.{col2});


                    DELETE
                    FROM {table} WHERE {fk} IN %(old)s;

                """
                ).format(**_params)
                cr.execute(_query, locals())

            if not col2:  # it's a model
                # update default values
                # TODO? update all defaults using 1 query (using `WHERE (model, name) IN ...`)
                model = model_of_table(cr, table)
                if table_exists(cr, "ir_values"):
                    _params = {
                        "cast0": sql.SQL(cast_write.partition("%s")[0]),
                        "cast2": sql.SQL(cast_write.partition("%s")[2]),
                        "column": sql.Identifier(column_read),
                    }

                    _query = sql.SQL(
                        """

                        UPDATE ir_values
                        SET value = {cast0} %(pmap)s::json->>({column}) {cast2}
                        WHERE KEY='default'
                          AND model = %(model)s
                          AND name = %(fk)s
                          AND {column} IN %(pmap_keys)s

                    """
                    ).format(**_params)
                    cr.execute(_query, locals())
                else:
                    _query = sql.SQL(
                        """

                        UPDATE ir_default d
                        SET json_value = %(smap)s::json->>json_value
                        FROM ir_model_fields f
                        WHERE f.id = d.field_id
                          AND model = %(model)s
                          AND name = %(fk)s
                          AND d.json_value IN %(pmap_keys)s

                    """
                    )
                    cr.execute(_query, locals())

    # indirect references
    for ir in indirect_references(cr, bound_only=True):
        if ir.table == "ir_model_data" and not replace_xmlid:
            continue
        _params = {}
        upd = sql.SQL("")
        if ir.res_model:
            _params["model"] = sql.Identifier(ir.res_model)
            upd += sql.SQL("{model} = %(model_dst)s,").format(**_params)
        if ir.res_model_id:
            _params["model_id"] = sql.Identifier(ir.res_model_id)
            upd += sql.SQL(
                "{model_id} = (SELECT id FROM ir_model WHERE model = %(model_dst)s),"
            ).format(**_params)
        where = sql.SQL(ir.model_filter(placeholder="%(model_src)s"))

        _params = dict(
            {
                "table": sql.Identifier(ir.table),
                "res_id": sql.Identifier(ir.res_id),
                "upd": upd,
                "where": where,
            },
            **_params
        )

        _query = sql.SQL(
            """

            UPDATE {table}
            SET {upd} {res_id} = (%(jmap)s::json->>{res_id}::varchar)::int4
            WHERE {where}
              AND {res_id} IN %(old)s

        """
        ).format(**_params)
        cr.execute(_query, locals())

    # reference fields
    cmap, cmap_keys = genmap("%s,%%d" % model_src, "%s,%%d" % model_dst)
    cr.execute("SELECT model, name FROM ir_model_fields WHERE ttype='reference'")
    for model, column in cr.fetchall():
        table = table_of_model(cr, model)
        if column_exists(cr, table, column):
            _params = {"table": sql.Identifier(table), "column": sql.Identifier(column)}

            _query = sql.SQL(
                """

                UPDATE {table}
                SET {column} = %(cmap)s::json->>{column}
                WHERE {column} IN %(cmap_keys)s

            """
            ).format(**_params)
            cr.execute(_query, locals())


# ---------- UI Utilities (Views, Menus) ----------
def _update_view_key(cr, old, new):
    """Update the key of a view."""
    if not column_exists(cr, "ir_ui_view", "key"):
        return

    _query = sql.SQL(
        """


        UPDATE ir_ui_view v
        SET key = CONCAT(%(new)s, '.', x.name)
        FROM ir_model_data x
        WHERE x.model = 'ir.ui.view'
          AND x.res_id = v.id
          AND x.module = %(old)s
          AND v.key = CONCAT(x.module, '.', x.name)

    """
    )
    cr.execute(_query, locals())


def remove_view(
    cr,
    xml_id=None,
    view_id=None,
    deactivate_custom=DROP_DEPRECATED_CUSTOM,
    silent=False,
):
    """
    Recursively delete the given view and its inherited views.

    Delete the view and its inherited views as long as they are part of a module.
    Will crash as soon as a custom view exists anywhere
    in the hierarchy.

        :param xml_id=None: fully qualified xmlid of the view
        :param view_id=None: id of the view
        :param deactivate_custom=False: if set, any custom view inheriting from
                                        any of the deleted views will be
                                        deactivated, otherwise a MigrationError
                                        will be raised if a custom view exists;
                                        can be set by the system environment
                                        variable OE_DROP_DEPRECATED_CUSTOM
        :param silent=False: if True, no log output will be generated

    Note that you can either provide an xml_id or view_id but not both.
    """
    assert bool(xml_id) ^ bool(view_id), "You Must specify either xmlid or view_id"
    if xml_id:
        view_id = ref(cr, xml_id)
        if not view_id:
            return

        module, _, name = xml_id.partition(".")
        _query = sql.SQL(
            """

            SELECT model
            FROM ir_model_data
            WHERE module = %(module)s
              AND name = %(name)s

        """
        )
        cr.execute(_query, locals())

        [model] = cr.fetchone()
        if model != "ir.ui.view":
            raise ValueError(
                "{!r} should point to a 'ir.ui.view', not a {!r}".format(xml_id, model)
            )
    elif not silent or deactivate_custom:
        # search matching xmlid for logging or renaming of custom views

        _query = sql.SQL(
            """

            SELECT module,
                   name
            FROM ir_model_data
            WHERE model='ir.ui.view'
              AND res_id = %(view_id)s

        """
        )
        cr.execute(_query, locals())
        if cr.rowcount:
            xml_id = "%s.%s" % cr.fetchone()
        else:
            xml_id = "?"

    _query = sql.SQL(
        """


        SELECT v.id,
               x.module || '.' || x.name
        FROM ir_ui_view v
        LEFT JOIN ir_model_data x ON ( v.id = x.res_id
                                      AND x.model = 'ir.ui.view'
                                      AND x.module !~ '^_' )
        WHERE v.inherit_id = %(view_id)s

    """
    )
    cr.execute(_query, locals())
    for child_id, child_xml_id in cr.fetchall():
        if child_xml_id:
            if not silent:
                _logger.info(
                    "Dropping deprecated built-in view %s (ID %s), "
                    "as parent %s (ID %s) is going to be removed",
                    child_xml_id,
                    child_id,
                    xml_id,
                    view_id,
                )
            remove_view(
                cr, child_xml_id, deactivate_custom=deactivate_custom, silent=True
            )
        else:
            if deactivate_custom:
                if not silent:
                    _logger.warning(
                        "Deactivating deprecated custom view with "
                        "ID %s, as parent %s (ID %s) was removed",
                        child_id,
                        xml_id,
                        view_id,
                    )
                disable_view_query = sql.SQL(
                    """

                    UPDATE ir_ui_view
                    SET name = (name || ' - old view, inherited from ' || %(xml_id)s),
                        model = (model || '.disabled'),
                        inherit_id = NULL
                    WHERE id = %(child_id)s

                """
                )
                cr.execute(disable_view_query, locals())
            else:
                raise MigrationError(
                    "Deprecated custom view with ID %s needs migration, "
                    "as parent %s (ID %s) is going to be removed"
                    % (child_id, xml_id, view_id)
                )
    if not silent:
        _logger.info("Dropping deprecated built-in view %s (ID %s).", xml_id, view_id)
    remove_record(cr, xml_id)


@contextmanager
def edit_view(cr, xmlid=None, view_id=None, skip_if_noupdate=False):
    """Contextmanager that may yield etree arch of a view.

    As it may not yield, you must use `skippable_cm`:

    with migration.skippable_cm(), migration.edit_view(cr, 'xmlid') as arch:
        arch.attrib['string'] = 'My Form'

        :param xml_id=None: fully qualified xmlid of the view
        :param view_id=None: id of the view
        :param skip_if_noupdate=False: if True, will not yield if the view is
                                       set to be non-updatable

    Note that you can either provide an xml_id or view_id but not both.
    """
    assert bool(xmlid) ^ bool(view_id), "You Must specify either xmlid or view_id"
    noupdate = True
    if xmlid:
        if "." not in xmlid:
            raise ValueError("Please use fully qualified name <module>.<name>")

        module, _, name = xmlid.partition(".")

        _query = sql.SQL(
            """

            SELECT res_id,
                   noupdate
            FROM ir_model_data
            WHERE module = %(module)s
              AND name = %(name)s

        """
        )
        cr.execute(_query, locals())
        data = cr.fetchone()
        if data:
            view_id, noupdate = data

    if view_id and not (skip_if_noupdate and noupdate):
        _query = sql.SQL(
            """

            SELECT arch_db
            FROM ir_ui_view
            WHERE id = %(view_id)s

        """
        )
        cr.execute(_query, locals())
        [arch] = cr.fetchone() or [None]
        if arch:
            arch = lxml.etree.fromstring(arch)
            yield arch
            arch = lxml.etree.tostring(arch, encoding="unicode")
            _query = sql.SQL(
                """

                UPDATE ir_ui_view
                SET arch_db = %(arch)s
                WHERE id = %(view_id)s

            """
            )
            cr.execute(_query, locals())


def remove_menus(cr, menu_ids):
    """Remove ir.ui.menu records with the provided ids (and their children)."""
    if not menu_ids:
        return
    menu_ids = tuple(menu_ids)
    _query = sql.SQL(
        """

         WITH RECURSIVE tree(id) AS
          ( SELECT id
           FROM ir_ui_menu
           WHERE id IN %(menu_ids)s
           UNION SELECT m.id
           FROM ir_ui_menu m
           JOIN tree t ON (m.parent_id = t.id) )
        DELETE
        FROM ir_ui_menu m USING tree t
        WHERE m.id = t.id RETURNING m.id

    """
    )
    cr.execute(_query, locals())
    ids = tuple(x[0] for x in cr.fetchall())
    if ids:
        _query = sql.SQL(
            """

            DELETE
            FROM ir_model_data
            WHERE model='ir.ui.menu'
              AND res_id IN %(ids)s

        """
        )
        cr.execute(_query, locals())


# ---------- Database/Postgres Utilities ----------


def dbuuid(cr):
    """Get the uuid of the current database.

    In the case of a duplicated database, return the original uuid."""

    _query = sql.SQL(
        """

        SELECT value
        FROM ir_config_parameter
        WHERE KEY IN ('database.uuid',
                      'origin.database.uuid')
        ORDER BY KEY DESC LIMIT 1

    """
    )
    cr.execute(_query, locals())
    return cr.fetchone()[0]


def has_enterprise():
    """Check if the current installation has enterprise addons availables or not."""
    return bool(
        get_module_path("web_enterprise", downloaded=False, display_warning=False)
    )


def model_of_table(cr, table):
    """Return the model name for the provided table."""
    return {
        # could also be ir.actions.act_window_close, but we have yet
        # to encounter a case where we need it
        "ir_actions": "ir.actions.actions",
        "ir_act_url": "ir.actions.act_url",
        "ir_act_window": "ir.actions.act_window",
        "ir_act_window_view": "ir.actions.act_window.view",
        "ir_act_client": "ir.actions.client",
        "ir_act_report_xml": "ir.actions.report",
        "ir_act_server": "ir.actions.server",
        "ir_act_wizard": "ir.actions.wizard",
        "ir_config_parameter": "ir.config_parameter",
    }.get(table, table.replace("_", "."))


def table_of_model(cr, model):
    """Return the table for the provided model name."""
    return {
        "ir.actions.actions": "ir_actions",
        "ir.actions.act_url": "ir_act_url",
        "ir.actions.act_window": "ir_act_window",
        "ir.actions.act_window_close": "ir_actions",
        "ir.actions.act_window.view": "ir_act_window_view",
        "ir.actions.client": "ir_act_client",
        "ir.actions.report.xml": "ir_act_report_xml",
        "ir.actions.report": "ir_act_report_xml",
        "ir.actions.server": "ir_act_server",
        "ir.actions.wizard": "ir_act_wizard",
    }.get(model, model.replace(".", "_"))


def table_exists(cr, table):
    """Check if the specified table exists."""
    _query = sql.SQL(
        """

        SELECT 1
        FROM information_schema.tables
        WHERE table_name = %(table)s
          AND table_type = 'BASE TABLE'

    """
    )
    cr.execute(_query, locals())
    return cr.fetchone() is not None


def column_exists(cr, table, column):
    """Check if the column exist on the specified table."""
    return column_type(cr, table, column) is not None


def column_type(cr, table, column):
    """Get the type of the column on the specified table."""
    _query = sql.SQL(
        """

        SELECT udt_name
        FROM information_schema.columns
        WHERE table_name = %(table)s
          AND column_name = %(column)s

    """
    )
    cr.execute(_query, locals())
    r = cr.fetchone()
    return r[0] if r else None


def create_column(cr, table, column, definition):
    """Create a column on the specified table.

        :param str table: name of the table on which the column will be created
        :param str column: name of the new column
        :param str definition: SQL-style definition of the table type
                               e.g. `boolean` or `varchar(256)`
    """
    curtype = column_type(cr, table, column)
    if curtype:
        # TODO compare with definition
        pass
    else:
        _params = {
            "table": sql.Identifier(table),
            "column": sql.Identifier(column),
            "definition": sql.SQL(definition),
        }
        _query = sql.SQL(
            """

            ALTER TABLE {table} ADD COLUMN {column} {definition}

        """
        ).format(**_params)
        cr.execute(_query, locals())


def remove_column(cr, table, column, cascade=False):
    """Remove a column.

        :param str table: name of the table on which the column will be dropped
        :param str column: name of the column to drop
        :param bool cascade: if True, all records having a FKEY pointing to this column
                             will be cascade-deleted (default: False)
    """
    if column_exists(cr, table, column):
        drop_depending_views(cr, table, column)
        _params = {
            "table": sql.Identifier(table),
            "column": sql.Identifier(column),
            "drop_cascade": sql.SQL("CASCADE" if cascade else ""),
        }
        _query = sql.SQL(
            """

            ALTER TABLE {table} DROP COLUMN {column} {drop_cascade}

        """
        ).format(**_params)
        cr.execute(_query, locals())


def get_columns(cr, table, ignore=("id",), extra_prefixes=None):
    """
    Get the list of column in a table, minus ignored ones.

    Can also returns the list multiple times with different prefixes.
    This can be used to duplicating records (INSERT SELECT from the same table)

        :param table: table toinspect
        :param ignore=('id'): tuple of column names to ignore
    """
    select = sql.SQL("quote_ident(column_name)")
    params = []
    if extra_prefixes:
        select = sql.SQL(",").join(
            [select]
            + [sql.SQL("concat(%s, '.', {select}").format(select)] * len(extra_prefixes)
        )
        params = list(extra_prefixes)

    _params = {"select": select}
    _query = sql.SQL(
        """

        SELECT {select}
        FROM information_schema.columns
        WHERE table_name = %s
          AND column_name NOT IN %s

    """
    ).format(**_params)
    cr.execute(_query, params + [table, ignore])  # Params is a list of unnamed args
    return list(zip(*cr.fetchall()))


def get_depending_views(cr, table, column):
    """Get the list of SQL views depending on the specified column."""
    # http://stackoverflow.com/a/11773226/75349
    _query = sql.SQL(
        """

        SELECT DISTINCT quote_ident(dependee.relname)
        FROM pg_depend
        JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid
        JOIN pg_class AS dependee ON pg_rewrite.ev_class = dependee.oid
        JOIN pg_class AS dependent ON pg_depend.refobjid = dependent.oid
        JOIN pg_attribute ON pg_depend.refobjid = pg_attribute.attrelid
        AND pg_depend.refobjsubid = pg_attribute.attnum
        WHERE dependent.relname = %(table)s
          AND pg_attribute.attnum > 0
          AND pg_attribute.attname = %(column)s
          AND dependee.relkind='v'

    """
    )
    cr.execute(_query, locals())
    return map(itemgetter(0), cr.fetchall())


def drop_depending_views(cr, table, column):
    """Drop views depending on a column.

    This is usually used to ensure that modifying fields will not make
    SQL views (auto=False models) crash after the upgrade, forcing them
    to be regenerated during the uprade.
    """
    for view in get_depending_views(cr, table, column):
        _params = {"view": sql.Identifier(view)}
        _query = sql.SQL(
            """

            DROP VIEW IF EXISTS {} CASCADE

        """
        ).format(**_params)
        cr.execute(_query, locals())


def get_fk(cr, table):
    """Get the list of foreign keys pointing to `table`

        :rtype: list(tuple)
        :return: [(foreign_table, foreign_column, constraint_name, on_delete_action)]
                 where on_delete_action if one of the following:
                 - a: no action
                 - r: restrict
                 - c: cascade
                 - n: set null
                 - d: set default
    """

    _query = sql.SQL(
        """

        SELECT quote_ident(cl1.relname) AS TABLE,
               quote_ident(att1.attname) AS COLUMN,
               quote_ident(con.conname) AS conname,
               con.confdeltype
        FROM pg_constraint AS con,
             pg_class AS cl1,
             pg_class AS cl2,
             pg_attribute AS att1,
             pg_attribute AS att2
        WHERE con.conrelid = cl1.oid
          AND con.confrelid = cl2.oid
          AND array_lower(con.conkey, 1) = 1
          AND con.conkey[1] = att1.attnum
          AND att1.attrelid = cl1.oid
          AND cl2.relname = %(table)s
          AND att2.attname = 'id'
          AND array_lower(con.confkey, 1) = 1
          AND con.confkey[1] = att2.attnum
          AND att2.attrelid = cl2.oid
          AND con.contype = 'f'

    """
    )
    cr.execute(_query, locals())
    return cr.fetchall()


def delete_unused(cr, table, xmlids, set_noupdate=True):
    """
    Delete all records in the provided list not being referenced in any FKEY.

    Note that the xmlids themselves are not removed.

        :param table: target table for the deletion
        :param xmlids: list of xmlids to be checked and potentially deleted
        :param set_noupdate=True: if set, the noupdate field of all the provided
                                  xmlids will be set to True
    """
    sub = sql.SQL(" UNION ").join(
        [
            sql.SQL(
                """

                SELECT 1
                FROM "{}" x
                WHERE x."{}" = t.id

            """.format(
                    f[0], f[1]
                )
            )
            for f in get_fk(cr, table)
        ]
    )
    idmap = {ref(cr, x): x for x in xmlids}
    idmap.pop(None, None)
    if not sub or not idmap:
        return
    idmap = list(idmap)
    _params = {"table": sql.Identifier(table), "subquery": sub}
    _query = sql.SQL(
        """

        SELECT id
        FROM {table} t
        WHERE id = ANY(%(idmap)s)
          AND NOT EXISTS({subquery})

    """
    ).format(**_params)
    cr.execute(_query, locals())

    for (tid,) in cr.fetchall():
        remove_record(cr, idmap.pop(tid))

    if set_noupdate:
        for xid in idmap.values():
            force_noupdate(cr, xid, True)


def get_index_on(cr, table, *columns):
    """Get the list of indexes on the provided column names.

        :rtype: list(tuple)
        :return:a [(index_name, unique, pk)]
    """
    columns = sorted(columns)
    _query = sql.SQL(
        """


        SELECT name,
               indisunique,
               indisprimary
        FROM
          (SELECT quote_ident(i.relname) AS name,
                  x.indisunique,
                  x.indisprimary,
                  array_agg(a.attname::text
                            ORDER BY a.attname) AS attrs
           FROM
             (SELECT *,
                     unnest(indkey) AS unnest_indkey
              FROM pg_index) x
           JOIN pg_class c ON c.oid = x.indrelid
           JOIN pg_class i ON i.oid = x.indexrelid
           JOIN pg_attribute a ON (a.attrelid=c.oid
                                   AND a.attnum=x.unnest_indkey)
           WHERE (c.relkind = ANY (ARRAY['r'::"char", 'm'::"char"]))
             AND i.relkind = 'i'::"char"
             AND c.relname = %(table)s
           GROUP BY 1,
                    2,
                    3 ) idx
        WHERE attrs = %(columns)s

    """
    )
    cr.execute(_query, locals())
    return cr.fetchone()


def pg_array_uniq(a, drop_null=False):
    """???"""
    dn = "WHERE x IS NOT NULL" if drop_null else ""
    return "ARRAY(SELECT x FROM unnest({}) x {} GROUP BY x)".format(a, dn)


def pg_html_escape(s, quote=True):
    """SQL version of html.escape"""
    replacements = [("&", "&amp;"), ("<", "&lt;"), (">", "&gt;")]  # Must be done first!
    if quote:
        replacements += [('"', "&quot;"), ("'", "&#x27;")]

    def q(s):
        return (
            psycopg2.extensions.QuotedString(s).getquoted().decode("utf-8")
        )  # noqa: E704

    return reduce(
        lambda s, r: "replace({}, {}, {})".format(s, q(r[0]), q(r[1])), replacements, s
    )


def pg_text2html(s):
    return r"CONCAT('<p>', replace({}, E'\n', '<br>'), '</p>')".format(
        pg_html_escape(s)
    )


def view_exists(cr, view):
    """Check if the specified SQL view exists."""

    _query = sql.SQL(
        """

        SELECT 1
        FROM information_schema.views
        WHERE TABLE_NAME=%(view)s

    """
    )
    cr.execute(_query, locals())
    return bool(cr.rowcount)


# ----------------- Utils -----------------


def remove_record(cr, name, deactivate=False, active_field="active"):
    """
    Remove a record from the database by xmlid.

        :param str name: fully qualified xmlid (<module>.<name>)
        :param boolean deactivate: if True, the record may be archived if
                                   deletion is impossible (eg FKEY constraint)
        :param str active_field: name of the field to use for deactivation,
                                 'active' by default
    """
    if "." not in name:
        raise ValueError("Please use fully qualified name <module>.<name>")
    module, name = name.split(".")

    _query = sql.SQL(
        """

        DELETE
        FROM ir_model_data
        WHERE module = %(module)s
          AND name = %(name)s RETURNING model, res_id

    """
    )
    cr.execute(_query, locals())
    data = cr.fetchone()
    if not data:
        return
    model, res_id = data
    table = table_of_model(cr, model)
    try:
        with savepoint(cr):
            cr.execute('DELETE FROM "%s" WHERE id=%%s' % table, (res_id,))
    except Exception:
        if not deactivate or not active_field:
            raise
        _params = {
            "table": sql.Identifier(table),
            "active_field": sql.Identifier(active_field),
        }
        _query = sql.SQL(
            """

            UPDATE {table}
            SET {active_field} = FALSE
            WHERE id = %(res_id)s

        """
        ).format(**_params)
        cr.execute(_query, locals())
    else:
        # delete all indirect references to the record (e.g. mail_message entries, etc.)
        for ir in indirect_references(cr, bound_only=True):
            _params = {
                "table": sql.Identifier(ir.table),
                "where_clause": sql.SQL(ir.model_filter(placeholder="%(model)s")),
                "res_id": sql.Identifier(ir.res_id),
            }
            _query = sql.SQL(
                """

                DELETE
                FROM {table}
                WHERE {where_clause}
                  AND {res_id} = %(res_id)s

            """
            ).format(**_params)
            cr.execute(_query, locals())


def splitlines(s):
    """Yield stripped lines of `s`.

    Skip empty lines & remove comments (starts with `#`).
    """
    return (sl for l in s.splitlines() for sl in [l.split("#", 1)[0].strip()] if sl)


def expand_braces(s):
    """Expand braces (a la bash).

    Only handle one expension of a 2 parts (because we don't need more).
    """
    r = re.compile(r"(.*){([^},]*?,[^},]*?)}(.*)")
    m = r.search(s)
    if not m:
        raise ValueError("No braces to expand")
    head, match, tail = m.groups()
    a, b = match.split(",")
    return [head + a + tail, head + b + tail]


class IndirectReference(
    namedtuple("IndirectReference", "table res_model res_id res_model_id")
):
    def model_filter(self, prefix="", placeholder="%s"):
        if prefix and prefix[-1] != ".":
            prefix += "."
        if self.res_model_id:
            placeholder = "(SELECT id FROM ir_model WHERE model={})".format(placeholder)
            column = self.res_model_id
        else:
            column = self.res_model

        return '{}"{}"={}'.format(prefix, column, placeholder)


# allow the class to handle defaults implicitely
IndirectReference.__new__.__defaults__ = (
    None,
    None,
)  # https://stackoverflow.com/a/18348004


def indirect_references(cr, bound_only=False):
    IR = IndirectReference
    each = [
        IR("ir_attachment", "res_model", "res_id"),
        IR("ir_cron", "model", None),
        IR("ir_act_report_xml", "model", None),
        IR("ir_act_window", "res_model", "res_id"),
        IR("ir_act_window", "src_model", None),
        IR("ir_act_server", "wkf_model_name", None),
        IR("ir_act_server", "crud_model_name", None),
        IR("ir_act_client", "res_model", None),
        IR("ir_model", "model", None),
        IR("ir_model_fields", "model", None),
        # destination of a relation field
        IR("ir_model_fields", "relation", None),
        IR("ir_model_data", "model", "res_id"),
        IR("ir_filters", "model_id", None),  # YUCK!, not an id
        IR("ir_exports", "resource", None),
        IR("ir_ui_view", "model", None),
        IR("ir_values", "model", "res_id"),
        IR("wkf_transition", "trigger_model", None),
        IR("wkf_triggers", "model", None),
        IR("ir_model_fields_anonymization", "model_name", None),
        IR("ir_model_fields_anonymization_migration_fix", "model_name", None),
        IR("base_import_import", "res_model", None),
        IR("calendar_event", "res_model", "res_id"),  # new in saas~18
        IR("mail_template", "model", None),
        IR("mail_activity", "res_model", "res_id", "res_model_id"),
        IR("mail_alias", None, "alias_force_thread_id", "alias_model_id"),
        IR("mail_alias", None, "alias_parent_thread_id", "alias_parent_model_id"),
        IR("mail_followers", "res_model", "res_id"),
        IR("mail_message_subtype", "res_model", None),
        IR("mail_message", "model", "res_id"),
        IR("mail_compose_message", "model", "res_id"),
        IR("mail_wizard_invite", "res_model", "res_id"),
        IR("mail_mail_statistics", "model", "res_id"),
        IR("mail_mass_mailing", "mailing_model", None),
        IR("project_project", "alias_model", None),
        IR("rating_rating", "res_model", "res_id", "res_model_id"),
        IR("rating_rating", "parent_res_model", "parent_res_id", "parent_res_model_id"),
    ]

    for ir in each:
        if bound_only and not ir.res_id:
            continue
        if ir.res_id and not column_exists(cr, ir.table, ir.res_id):
            continue

        # some `res_model/res_model_id` combination may change between
        # versions (i.e. rating_rating.res_model_id was added in saas~15).
        # we need to verify existance of columns before using them.
        if ir.res_model and not column_exists(cr, ir.table, ir.res_model):
            ir = ir._replace(res_model=None)
        if ir.res_model_id and not column_exists(cr, ir.table, ir.res_model_id):
            ir = ir._replace(res_model_id=None)
        if not ir.res_model and not ir.res_model_id:
            continue

        yield ir


def res_model_res_id(cr):
    """Iterate on base models having a field that references records by model/id.

    Allow iterating on all models that reference records by using a
    res_model/res_id field combiination of similar reference system;
    usually used to quickly iterate over basic ORM models (views, crons, etc.)
    that references other models in an indirect way (without FKEY).

        :rtype tuple:
        :return: tuple containing the model, model reference field name
                 and id reference field name
    """
    for ir in indirect_references(cr):
        if ir.res_model:
            yield model_of_table(cr, ir.table), ir.res_model, ir.res_id


@contextmanager
def skippable_cm():
    """Allow a contextmanager to not yield."""
    if not hasattr(skippable_cm, "_msg"):

        @contextmanager
        def _():
            if 0:
                yield

        try:
            with _():
                pass
        except RuntimeError as r:
            skippable_cm._msg = str(r)
    try:
        yield
    except RuntimeError as r:
        if str(r) != skippable_cm._msg:
            raise


@contextmanager
def savepoint(cr):
    """Provide a savepoint context.

    If a query executed in this context fails, the operation is rollbacked,
    otherwise it success silently.
    """
    name = hex(int(time.time() * 1000))[1:]
    cr.execute("SAVEPOINT {}".format(name))
    try:
        yield
        cr.execute("RELEASE SAVEPOINT {}".format(name))
    except Exception:
        cr.execute("ROLLBACK TO SAVEPOINT {}".format(name))
        raise


def log_progress(it, qualifier="elements", logger=_logger, size=None):
    if size is None:
        size = len(it)
    size = float(size)
    t0 = t1 = datetime.datetime.now()
    for i, e in enumerate(it, 1):
        yield e
        t2 = datetime.datetime.now()
        if (t2 - t1).total_seconds() > 60:
            t1 = datetime.datetime.now()
            tdiff = t2 - t0
            logger.info(
                "[%.02f%%] %d/%d %s processed in %s (TOTAL estimated time: %s)",
                (i / size * 100.0),
                i,
                size,
                qualifier,
                tdiff,
                datetime.timedelta(seconds=tdiff.total_seconds() * size / i),
            )


def env(cr):
    """Get an environment for the SUPERUSER ('admin')."""
    from odoo.api import Environment

    return Environment(cr, SUPERUSER_ID, {})


def import_script(path):
    """Import a script from another module.

    This can be used if some changes have been applied across the whole
    codebase but would need module-specific changes; e.g. a generic script
    can be defined in the `base` module and called by any relevant module
    who might want to use it to adapt something depending on values depending
    from that module.

        :param path: relative path to the script
                     <module>/<versions>/<script>
                     e.g. base/12.0.1.3/do_stuff.py
    """
    name, _ = os.path.splitext(os.path.basename(path))
    full_path = os.path.join(os.path.dirname(__file__), path)
    with open(full_path) as fp:
        return imp.load_source(name, full_path, fp)


def dispatch_by_dbuuid(cr, version, callbacks):
    """Apply dbuuid-specific migrations.

    Allow defining custom migration functions that can be applied on a
    single database identified by its uuid, e.g.:

    _db_callback(cr, version):
        # do stuff specific to that db

    def migrate(cr, version):
        migration.dispatch_by_dbuuid(cr, version, {
            '88ef269b-f6de-4f76-b9c8-868fc5569136': _db_callback,
        })

    The callback function _db_callback should have the same signature as a
    `migrate` function, taking the cursor `cr` and `version` as args.

        :param version: target version for this upgrade
        :param callbacks: dict where each key is a uuid and value is a
                          reference to the callback function
    """
    uuid = dbuuid(cr)
    if uuid in callbacks:
        func = callbacks[uuid]
        _logger.info("calling dbuuid-specific function `%s`", func.__name__)
        func(cr, version)


def register_unanonymization_query(
    cr, model, field, query, query_type="sql", sequence=10
):
    """
    Generate an unanonymization query.

    Allow newly created fields, values, etc. to be deanonymzed even though they
    were not present during anonymization (e.g. when values are moved/copied).

        :param model:
        :param field:
        :param query:
        :param query_type='sql':
        :param sequence=10:
    """
    target_version = release.major_version
    _query = sql.SQL(
        """
        INSERT INTO ir_model_fields_anonymization_migration_fix( target_version, sequence, query_type, model_name, field_name, query )
        VALUES (%(target_version)s,
                %(sequence)s,
                %(query_type)s,
                %(model)s,
                %(field)s,
                %(query)s)

    """
    )
    cr.execute(_query, locals())


def _rst2html(rst):
    """Convert rst to html."""
    overrides = dict(
        embed_stylesheet=False,
        doctitle_xform=False,
        output_encoding="unicode",
        xml_declaration=False,
    )
    html = publish_string(
        source=dedent(rst), settings_overrides=overrides, writer=MyWriter()
    )
    return html_sanitize(html, silent=False)


def _md2html(md):
    """Convert markdown to html."""
    extensions = [
        "markdown.extensions.smart_strong",
        "markdown.extensions.nl2br",
        "markdown.extensions.sane_lists",
    ]
    return markdown.markdown(md, extensions=extensions)


# ---------- xmlid utilities ----------


def ref(cr, xmlid):
    """Get the id of an xmlid entry."""
    if "." not in xmlid:
        raise ValueError("Please use fully qualified name <module>.<name>")

    module, name = xmlid.split(".")

    _query = sql.SQL(
        """

        SELECT res_id
        FROM ir_model_data
        WHERE module = %(module)s
          AND name = %(name)s

    """
    )
    cr.execute(_query, locals())
    data = cr.fetchone()
    if data:
        return data[0]
    return None


def rename_xmlid(cr, old, new, noupdate=None):
    """Rename an xmlid.

    In the case of a view xmlid, the key if the view is updated as well.
    """
    if "." not in old or "." not in new:
        raise ValueError("Please use fully qualified name <module>.<name>")

    old_module, old_name = old.split(".")
    new_module, new_name = new.split(".")
    noupdate = sql.SQL(
        "" if noupdate is None else (", noupdate=" + str(bool(noupdate)).lower())
    )
    _params = {"noupdate": noupdate}
    _query = sql.SQL(
        """

        UPDATE ir_model_data
        SET module = %(new_module)s,
            name=%(new_name)s {noupdate}
        WHERE module = %(old_module)s
          AND name = %(old_name)s RETURNING model,
                      res_id

    """
    ).format(**_params)
    cr.execute(_query, locals())
    data = cr.fetchone()
    if data:
        model, rid = data
        if model == "ir.ui.view":
            _update_view_key(cr, old, new)
            _query = sql.SQL(
                """

                UPDATE ir_ui_view
                SET key = %(new)s
                WHERE id = %(rid)s
                  AND key = %(old)s

            """
            ).format(**_params)
            cr.execute(_query, locals())
        return rid
    return None


def force_noupdate(cr, xmlid, noupdate=True, warn=False):
    """Force the noupdate value of an xmlid."""
    if "." not in xmlid:
        raise ValueError("Please use fully qualified name <module>.<name>")

    module, name = xmlid.split(".")
    _query = sql.SQL(
        """

        UPDATE ir_model_data
        SET noupdate = %(noupdate)s
        WHERE module = %(module)s
          AND name = %(name)s
          AND noupdate != %(noupdate)s

    """
    )
    cr.execute(_query, locals())
    if noupdate is False and cr.rowcount and warn:
        _logger.warning("Customizations on `%s` might be lost!", xmlid)
    return cr.rowcount


def ensure_xmlid_match_record(cr, xmlid, model, values):
    """Ensure the provided values matches the provided xmlid.

    Check the provided model table for the presence of a record matching
    the given xmlid. If the record is not found, searches the table for
    any record matching the provided values and associate it to the xmlid.

        :param str xmlid: fully qualified xmlid (<module>.<name>)
        :param str model: name of the model to check (Odoo model name)
        :param tuple values: column name and value to check if the record needs
                             to be found in the table (will be used in a WHERE
                             query)
        :rtype: integer
        :return: ID of the record
    """
    if "." not in xmlid:
        raise ValueError("Please use fully qualified name <module>.<name>")

    module, name = xmlid.split(".")
    _query = sql.SQL(
        """

        SELECT id,
               res_id
        FROM ir_model_data
        WHERE module = %(module)s
          AND name = %(name)s

    """
    )
    cr.execute(_query, locals())

    table = table_of_model(cr, model)
    data = cr.fetchone()
    if data:
        data_id, res_id = data
        # check that record still exists
        _params = {"table": sql.Identifier(table)}
        _query = sql.SQL(
            """
            SELECT id
            FROM {table}
            WHERE id = %(res_id)s

        """
        ).format(**_params)
        cr.execute(_query, locals())
        if cr.fetchone():
            return res_id
    else:
        data_id = None

    # search for existing record marching values
    where = sql.Composed([])
    data = ()
    for field, value in values.items():
        _params = {"field": sql.Identifier(field)}
        if value:
            _params["value"] = sql.Literal(value)
            where += sql.Composed([sql.SQL("{field} = {value}").format(**_params)])
            data += (value,)
        else:
            where += sql.Composed([sql.SQL("{field} IS NULL").format(**_params)])
            data += ()

    _params = {"table": sql.Identifier(table), "where": where.join(" AND ")}
    _query = sql.SQL(
        """
        SELECT id
        FROM {table}
        WHERE {where}

    """
    ).format(**_params)
    cr.execute(_query, locals())
    record = cr.fetchone()
    if not record:
        return None

    res_id = record[0]

    # update xmlid table
    if data_id:
        _query = sql.SQL(
            """

            UPDATE ir_model_data
            SET res_id = %(res_id)s
            WHERE id = %(data_id)s

        """
        )
        cr.execute(_query, locals())
    else:

        _query = sql.SQL(
            """

            INSERT INTO ir_model_data (module, name, model, res_id, noupdate)
            VALUES (%(module)s,
                    %(name)s,
                    %(model)s,
                    %(res_id)s,
                    TRUE)

        """
        )
        cr.execute(_query, locals())
    return res_id


# -------- Announcement Message --------


_DEFAULT_HEADER = """
<p>{module} has been upgraded to version {version}.</p>
<h2>What's new in this upgrade?</h2>
"""

_DEFAULT_FOOTER = "<p>Enjoy this new version of {module}!</p>"

_DEFAULT_RECIPIENT = "mail.channel_all_employees"


def announce(
    cr,
    module,
    version,
    msg,
    format="rst",
    recipient=_DEFAULT_RECIPIENT,
    header=_DEFAULT_HEADER,
    footer=_DEFAULT_FOOTER,
    pluses_for_enterprise=None,
):
    """
    Post an upgrade message in the selected channel detailing the upgrade.

        :param module: nmae of the upgraded module
        :param version: target upgrade version
        :param msg: message regarding the upgrade
        :param format='rst': format of the message
                             ('rst' for ReStructured Text or 'md' for markdown)
        :param recipient: xmlid of the channel where the message will be posted
        :param header: header of the message (set False for no header)
        :param footer: footer of the message (set False for no footer)
        :param pluses_for_enterprise=None: if True, list elements in your message
                                           prefixed with a '+ ' string will be
                                           filtered out if the upgraded database
                                           does not have the Enterprise edition
    """
    if pluses_for_enterprise:
        plus_re = r"^(\s*)\+ (.+)\n"
        replacement = r"\1- \2\n" if has_enterprise() else ""
        msg = re.sub(plus_re, replacement, msg, flags=re.M)

    # do not notify early, in case the migration fails halfway through
    ctx = {"mail_notify_force_send": False, "mail_notify_author": True}

    try:
        registry = env(cr)
        user = registry["res.users"].browse([SUPERUSER_ID])[0].with_context(ctx)

        def ref(xid):
            return registry.ref(xid).with_context(ctx)

    except MigrationError:
        registry = Registry.get(cr.dbname)
        user = registry["res.users"].browse(cr, SUPERUSER_ID, SUPERUSER_ID, context=ctx)

        def ref(xid):
            rmod, _, rxid = recipient.partition(".")
            return registry["ir.model.data"].get_object(
                cr, SUPERUSER_ID, rmod, rxid, context=ctx
            )

    # default recipient
    poster = user.message_post

    if recipient:
        try:
            poster = ref(recipient).message_post
        except (ValueError, AttributeError):
            # Cannot find record, post the message on the wall of the admin
            pass

    if format == "rst":
        msg = _rst2html(msg)
    elif format == "md":
        msg = _md2html(msg)

    message = ((header or "") + msg + (footer or "")).format(
        module=module or "Odoo", version=version
    )
    _logger.debug(message)

    type_field = "message_type"
    kw = {type_field: "notification"}

    try:
        poster(
            body=message,
            partner_ids=[user.partner_id.id],
            subtype="mail.mt_comment",
            **kw
        )
    except Exception:
        _logger.warning("Cannot announce message", exc_info=True)


# --- NOT SURE IF STILL NEEDED ??? --- #


def main(func, version=None):
    """a main() function for scripts"""
    # NOTE: this is not recommanded when the func callback use the ORM as the addon-path is
    # incomplete. Please pipe your script into `odoo shell`.
    # Do not forget to commit the cursor at the end.
    if len(sys.argv) != 2:
        sys.exit("Usage: {} <dbname>".format(sys.argv[0]))
    dbname = sys.argv[1]
    with db_connect(dbname).cursor() as cr, Environment.manage():
        func(cr, version)


def _ir_values_value(cr):
    # returns the casting from bytea to text needed in saas~17 for column `value` of `ir_values`
    # returns tuple(column_read, cast_write)
    result = getattr(_ir_values_value, "result", None)

    if result is None:
        if column_type(cr, "ir_values", "value") == "bytea":
            cr.execute(
                "SELECT character_set_name FROM information_schema.character_sets"
            )
            charset, = cr.fetchone()
            column_read = "convert_from(value, '%s')" % charset
            cast_write = "convert_to(%%s, '%s')" % charset
        else:
            column_read = "value"
            cast_write = "%s"
        _ir_values_value.result = result = (column_read, cast_write)

    return result


def chunks(iterable, size, fmt=None):
    """
    Split `iterable` into chunks of `size` and wrap each chunk
    using function 'fmt' (`iter` by default; join strings)

    >>> list(chunks(range(10), 4, fmt=tuple))
    [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9)]
    >>> ' '.join(chunks('abcdefghijklm', 3))
    'abc def ghi jkl m'
    >>>

    """
    if fmt is None:
        fmt = "".join

    it = iter(iterable)
    try:
        while True:
            yield fmt(chain((next(it),), islice(it, size - 1)))
    except StopIteration:
        return


def iter_browse(model, *args, **kw):
    """
    Iterate and browse through record without filling the cache.
    `args` can be `cr, uid, ids` or just `ids` depending on kind of `model` (old/new api)
    """
    assert len(args) in [1, 3]  # either (cr, uid, ids) or (ids,)
    cr_uid = args[:-1]
    ids = args[-1]
    chunk_size = kw.pop("chunk_size", 200)  # keyword-only argument
    logger = kw.pop("logger", _logger)
    if kw:
        raise TypeError("Unknow arguments: %s" % ", ".join(kw))

    def browse(ids):
        model.invalidate_cache(*cr_uid)
        args = cr_uid + (list(ids),)
        return model.browse(*args)

    def end():
        model.invalidate_cache(*cr_uid)
        if 0:
            yield

    it = chain.from_iterable(chunks(ids, chunk_size, fmt=browse))
    if logger:
        it = log_progress(it, qualifier=model._name, logger=logger, size=len(ids))

    return chain(it, end())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the dodoo-migrator (R) project.
# Copyright (c) 2018 Camptocamp SA and XOE Corp. SAS
# Authors: Guewen Baconnier, Leonardo Pistone, David Arnold, et al.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess

import pytest
from click.testing import CliRunner
from dodoo import odoo

from dodoo_migrator.cli import migrate

HERE = os.path.dirname(__file__)
DATADIR = os.path.join(HERE, "data/test_odoo_migration/")
MIG_TABLE = "dodoo_migrator"
MANIFEST_NAMES = ("__manifest__.py", "__openerp__.py")


def _exec_query(dbname, query):
    process = subprocess.Popen(
        ["psql", "-d", dbname, "-t", "-c", query], stdout=subprocess.PIPE
    )
    out, _ = process.communicate()
    return out


def test_all_queries_without_validating_result(odoodb, odoocfg):
    """ Test if all queries are syntax error free. """
    if odoo.release.version_info[0] != 12:
        pytest.skip("version <= 12 all modules / names where taken from v12.")
    # Test install, upgrade
    result = CliRunner().invoke(
        migrate,
        [
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.4.0-test-all-queries-syntax.yaml",
        ],
    )
    assert result.exit_code == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the dodoo-migrator (R) project.
# Copyright (c) 2018 Camptocamp SA and XOE Corp. SAS
# Authors: Guewen Baconnier, Leonardo Pistone, David Arnold, et al.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

import logging
import threading
import time

import click
import dodoo
import semver

from . import migration

_logger = logging.getLogger(__name__)


# The number below has been generated as below:
# pg_lock accepts an int8 so we build an hash composed with
# contextual information and we throw away some bits
#     lock_name = 'marabunta'
#     hasher = hashlib.sha1()
#     hasher.update('{}'.format(lock_name))
#     lock_ident = struct.unpack('q', hasher.digest()[:8])
# we just need an integer
ADVISORY_LOCK_IDENT = 7141416871301361999

MIGRATION_SCRIPTS_PATH = None


def get_additional_mig_path():
    return MIGRATION_SCRIPTS_PATH


def pg_advisory_lock(cursor, lock_ident):
    cursor.execute("SELECT pg_try_advisory_xact_lock(%s);", (lock_ident,))
    acquired = cursor.fetchone()[0]
    return acquired


class ApplicationLock(threading.Thread):
    def __init__(self, cr):
        self.acquired = False
        self.cr = cr
        self.replica = False
        self.stop = False
        super(ApplicationLock, self).__init__()

    def run(self):
        # If the migration is run concurrently (in several
        # containers, hosts, ...), only 1 is allowed to proceed
        # with the migration. It will be the first one to win
        # the advisory lock. The others will be flagged as 'replica'.
        while not pg_advisory_lock(self.cr, ADVISORY_LOCK_IDENT):
            if not self.replica:  # print only the first time
                _logger.WARN("A concurrent process is already " "running the migration")
            self.replica = True
            time.sleep(0.5)
        else:
            self.acquired = True
            idx = 0
            while not self.stop:
                # keep the connection alive to maintain the advisory
                # lock by running a query every 30 seconds
                if idx == 60:
                    self.cr.execute("SELECT 1")
                    idx = 0
                idx += 1
                # keep the sleep small to be able to exit quickly
                # when 'stop' is set to True
                time.sleep(0.5)


def do_migrate(env, file, since, until):
    """Perform a migration according to file.

    :param env: The odoo environment
    :type env: odoo.api.Environment
    :param file: The migration file to be applied
    :type config: file
    :param since: Migrate from this version onwards
    :type since: Version
    :param until: Migrate up to this version
    :type until: Version
    """

    with env.registry.cursor() as lock_connection:
        lock = ApplicationLock(lock_connection)
        lock.start()

        while not lock.acquired:
            time.sleep(0.5)
        else:
            if lock.replica:
                # when a replica could finally acquire a lock, it
                # means that the main process has finished the
                # migration. In that case, the replica should just
                # exit because the migration already took place. We
                # wait till then to be sure we won't run Odoo before
                # the main process could finish the migration.
                lock.stop = True
                lock.join()
                return
            # we are not in the replica: go on for the migration

        try:
            mig_spec = migration.MigrationSpec(env, file, since, until)
            mig_spec.run()
        finally:
            lock.stop = True
            lock.join()


@click.command(cls=dodoo.CommandWithOdooEnv)
@dodoo.options.addons_path_opt(True)
@dodoo.options.db_opt(True)
@click.option(
    "--file",
    "-f",
    default=".migrations.yaml",
    show_default=True,
    type=click.File("rb", lazy=True),
    help="The yaml file containing the migration steps.",
)
@click.option(
    "--mig-directory",
    "-m",
    type=click.Path(exists=True, file_okay=False),
    help="A migration directory shim. Layout after Odoo's migration"
    "folders within their named module folders."
    "Tipp: Can supply base migration scripts.",
)
@click.option(
    "--since",
    type=semver.VersionInfo.parse,
    required=False,
    help="Specify the version (excluded), to start from. If not "
    "specified, start from the latest applied version onwards.",
)
@click.option(
    "--until",
    type=semver.VersionInfo.parse,
    required=False,
    help="Specify the the target version, to which to migrate. If "
    "not specified, migrate up to the latest version.",
)
@click.option(
    "--metrics/--no-metrics",
    default=False,
    show_default=True,
    help="Prometheus metrics endpoint for migration progress. "
    "Can be consumed by a status page or monitoring solution.",
)
def migrate(env, file, mig_directory, since, until, metrics):
    """ Apply migration paths specified by a descriptive yaml migration file.

    Persists applied migrations within the target database.

    Connects to Odoo SA's migration service and can be run idempotently to
    check for results. Before uploading, can apply special before-steps. Once
    results are avialable, proceeds with remaining migration steps as specified
    by the migration file.

    A prometheus metrics endpoint is instrumented into the script. This can be
    scraped by a monitoring solution or a status page.
    """

    global MIGRATION_SCRIPTS_PATH
    MIGRATION_SCRIPTS_PATH = mig_directory
    do_migrate(env, file, since, until)


if __name__ == "__main__":  # pragma: no cover
    migrate()
